/*
 *      cdsdgui.c
 *      
 *      Copyright 2011 Alex <alex@linuxonly.ru>
 *      
 *      This program is free software: you can redistribute it and/or modify
 *      it under the terms of the GNU General Public License as published by
 *      the Free Software Foundation; either version 3 of the License, or
 *      (at your option) any later version.
 *      
 *      This program is distributed in the hope that it will be useful,
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of
 *      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *      GNU General Public License for more details.
 *      
 *      You should have received a copy of the GNU General Public License
 *      along with this program. If not, see <http://www.gnu.org/licenses/>.
 */

#include <gtk/gtk.h>
#include <glib/gi18n.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <dlfcn.h>

#include "service/libcdsd.h"
#include "service/resources.h"

#define CDSDGUI_GETTEXT_PACKAGE "cdsdgui"

#define CDSD_CD_SPEEDS_NUM   15
#define CDSD_DVD_SPEEDS_NUM  14
#define CDSD_BD_SPEEDS_NUM   9 

int cdspeeds[CDSD_CD_SPEEDS_NUM] = {0, 1, 2, 4, 8, 10, 12, 20, 32, 36, 40, 48, 52, 56, -1};
int dvdspeeds[CDSD_DVD_SPEEDS_NUM] = {0, 1, 2, 4, 6, 8, 10, 12, 16, 18, 20, 22, 24, -1};
int bdspeeds[CDSD_BD_SPEEDS_NUM] = {0, 1, 2, 4, 6, 8, 10, 12, -1};

#define CDSD_SPEED_TRAY_PACK(x, y)  ( (gint)((((x) << 16) & 0xFFFF0000) | ((y) & 0x0000FFFF)) )
#define CDSD_SPEED_TRAY_UNPACK1(x)  ( (gint)(((x) & 0xFFFF0000) >> 16) )
#define CDSD_SPEED_TRAY_UNPACK2(x)  ( (gint)((x) & 0x0000FFFF) )

enum {
	DEVS_COL_ICON = 0,
	DEVS_COL_ICON_SIZE,
	DEVS_COL_NAME,
	DEVS_COL_ID,
	DEVS_COL_CDSPEED,
	DEVS_COL_DVDSPEED,
	DEVS_COL_BDSPEED,
	DEVS_COL_POLLING,
	DEVS_COLUMNS
};

enum {
	CONTROLS_CD_SPEED = 0,
	CONTROLS_DVD_SPEED,
	CONTROLS_BD_SPEED,
	CONTROLS_POLLING
};

enum {
	INSTANCE_WINDOW = 0,
	INSTANCE_TRAY
};

enum {
	TRAY_CDSPEED = 0,
	TRAY_DVDSPEED,
	TRAY_BDSPEED
};

enum {
	INFO_DESCRIPTION = 0,
	INFO_VALUE,
	INFO_ID,
	INFO_COLUMNS
};

enum {
	INFO_DATA_STRING = 0,
	INFO_DATA_BYTE,
	INFO_DATA_SHORT,
	INFO_DATA_INT
};

typedef struct {
	struct cdsd_settings_device *device;
	GtkWidget *dialog;
} cdsd_longterm_command;

typedef struct {
	const gchar *description;
	const gchar *format;
	const gint datatype;
	void *value;
} cdsd_info_table;

struct cdsd_info_timer_data {
	cdsd_info_table *infotable;
	struct cdsd_settings_info *info;
	GtkListStore *store;
};

int  (*libcdsd_command_list_dl)(int *error, libcdsd_list_cb list_cb);
int  (*libcdsd_command_modeset_dl)(int *error, struct cdsd_settings_device *device);
int  (*libcdsd_command_info_dl)(int *error, int deviceid, struct cdsd_settings_info *info);
int  (*libcdsd_command_capabilities_dl)(int *error, int deviceid, struct cdsd_capabilities_info *capabilities);
int  (*libcdsd_connection_available_dl)(int *error);
/*int  (*libcdsd_get_error_dl)(void);*/


struct scdsdwindow {
	gboolean visible;
	GtkWidget *win;
	gchar *title;
	GtkWidget *menubar;
	GtkAccelGroup *accel_group;
	GtkWidget *devmenu, *helpmenu;
	GtkWidget *dev_menu, *apply_menu, *info_menu, *capabilities_menu, *sep_menu, *quit_menu, *help_menu, *readhelp_menu, *about_menu;
	GtkStatusIcon *trayicon;
	gulong traysigid;
	gboolean traylock;
	GtkWidget *traymenu, *devicesmenu, *speedsmenu, *cdspeeds_tm, *dvdspeeds_tm, *bdspeeds_tm, *cdspeedsmenu, *dvdspeedsmenu, *bdspeedsmenu, 
				*showwin_tm, *sep1_tm, *devices_tm, *sep2_tm, *quit_tm;
	GtkWidget *cdspeed_tm[CDSD_CD_SPEEDS_NUM - 1], *dvdspeed_tm[CDSD_DVD_SPEEDS_NUM - 1], *bdspeed_tm[CDSD_BD_SPEEDS_NUM - 1];
	GtkWidget *caption;
	GtkWidget *close;
	GtkWidget *apply;
	GtkWidget *list;
	GtkTreeSelection *selection;
	GtkWidget *table;
	GtkWidget *cdspeed;
	GtkWidget *dvdspeed;
	GtkWidget *bdspeed;
	GtkWidget *polling;
	gint devid, traydevid;
	gint cdspeedindex;
	gint dvdspeedindex;
	gint bdspeedindex;
	gint pollingindex;
};
struct scdsdwindow cdsdwindow;

static gint cdsdwindow_close(GtkWidget *widget, GdkEvent *event, gpointer data);
static void cdsdwindow_message(gchar *message, gchar *caption, gint type, GtkWidget *parent);
static gint cdsdwindow_about(GtkWidget *widget, GdkEvent *event, gpointer data);
static gint cdsdwindow_readhelp(GtkWidget *widget, GdkEvent *event, gpointer data);
static gboolean cdsdwindow_info_timer_foreach(GtkTreeModel *model, GtkTreePath *path, GtkTreeIter *iter, gpointer data);
static gboolean cdsdwindow_info_timer(gpointer data);
static gint cdsdwindow_info(GtkWidget *widget, GdkEvent *event, gpointer data);
static gint cdsdwindow_capabilities(GtkWidget *widget, GdkEvent *event, gpointer data);
static gboolean cdsdwindow_wait_dialog_timer(gpointer data);
static gboolean cdsdwindow_wait_dialog(gpointer data);
static gint cdsdwindow_apply(GtkWidget *widget, GdkEvent *event, gpointer data);
static void cdsdwindow_change_parameter(GtkComboBox *widget, gpointer data);
static void cdsdwindow_init_list(GtkWidget *list);
static gboolean cdsdwindow_set_controls(gint control, gint value, gint instance);
static gboolean cdsdwindow_device_set_speed_tray_foreach(GtkTreeModel *model, GtkTreePath *path, GtkTreeIter *iter, gpointer data);
static void cdsdwindow_device_set_speed_tray(GtkMenuItem *menuitem, gpointer data);
static gboolean cdsdwindow_device_selected_tray_foreach(GtkTreeModel *model, GtkTreePath *path, GtkTreeIter *iter, gpointer data);
static void cdsdwindow_device_selected_tray(GtkMenuItem *menuitem, gpointer data);
static int cdsdwindow_add_event_cb(struct cdsd_settings_device *device);
static void cdsdwindow_device_selected(GtkTreeView *tree_view, gpointer data);
static void cdsdwindow_set_action_buttons_state(gboolean enabled, gboolean changed);
static gint cdsdwindow_show(GtkWidget *widget, GdkEvent *event, gpointer data);
static void cdsdwindow_trayaction(GtkStatusIcon *status_icon, gpointer data);
static void cdsdwindow_traypopup(GtkStatusIcon *status_icon, guint button, guint activate_time, gpointer data);
static GtkWidget *cdsdwindow_create(void);
static int cdsdctl_load_library(void);
static void cdsdwindow_init_lib(void);
static void cdsdwindow_request_device_list(void);
#if GTK_CHECK_VERSION(3, 0, 0)
static void cdsdwindow_activate(GtkApplication *app);
#endif


static gint cdsdwindow_close(GtkWidget *widget, GdkEvent *event, gpointer data)
{
	cdsdwindow.visible = FALSE;
	gtk_widget_destroy(cdsdwindow.win);
	gtk_main_quit();
	
	return FALSE;
}

static void cdsdwindow_message(gchar *message, gchar *caption, gint type, GtkWidget *parent)
{
	GtkWidget *dialog;
	
	dialog = gtk_message_dialog_new((parent == NULL) ? NULL : GTK_WINDOW(parent),
									GTK_DIALOG_MODAL,
									type,
									GTK_BUTTONS_OK,
									message);
	
	gtk_window_set_title(GTK_WINDOW(dialog), caption);
	gtk_dialog_run(GTK_DIALOG(dialog));
	gtk_widget_destroy(dialog);
	
}

static gint cdsdwindow_about(GtkWidget *widget, GdkEvent *event, gpointer data)
{
	GtkWidget *dialog;
	GdkPixbuf *logo;
	GError *error = NULL;
	
	const gchar *authors[] = {
		"Alex <alex@linuxonly.ru>",
		NULL
	};
	
	const gchar *documenters[] = {
		"Alex <alex@linuxonly.ru>",
		NULL
	};
	
	const gchar *artists[] = {
		"Silvestre Herrera <silvestre.herrera@gmail.com>",
		NULL
	};
	
	const gchar *license = 
		"This program is free software: you can redistribute it and/or modify\n"
		"it under the terms of the GNU General Public License as published by\n"
		"the Free Software Foundation, either version 3 of the License, or\n"
		"(at your option) any later version.\n"
		"\n"
		"This program is distributed in the hope that it will be useful,\n"
		"but WITHOUT ANY WARRANTY; without even the implied warranty of\n"
		"MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the\n"
		"GNU General Public License for more details.\n"
		"\n"
		"You should have received a copy of the GNU General Public License\n"
		"along with this program.  If not, see <http://www.gnu.org/licenses/>.";
	
	dialog = gtk_about_dialog_new();
	logo = gdk_pixbuf_new_from_file(RESOURCE_MAIN_ICON, &error);
	if (error == NULL) {
		gtk_about_dialog_set_logo(GTK_ABOUT_DIALOG (dialog), logo);
	} else {
		if (error->domain == GDK_PIXBUF_ERROR) {
			g_print("GdkPixbufError: %s\n", error->message);
		} else if (error->domain == G_FILE_ERROR) {
			g_print("GFileError: %s\n", error->message);
		} else {
			g_print("An error in the domain: %d has occurred!\n", error->domain);
		}
		g_error_free(error);
	}
	
	#if GTK_CHECK_VERSION(2, 12, 0)
		gtk_about_dialog_set_program_name(GTK_ABOUT_DIALOG(dialog), "CDSD GUI");
	#else
		gtk_about_dialog_set_name(GTK_ABOUT_DIALOG(dialog), "CDSD GUI");
	#endif
	
	gtk_about_dialog_set_version(GTK_ABOUT_DIALOG(dialog), "0.0.1");
	gtk_about_dialog_set_copyright(GTK_ABOUT_DIALOG(dialog), "(C) 2011 Authors");
	gtk_about_dialog_set_comments(GTK_ABOUT_DIALOG(dialog), _("CDSD control"));
	
	gtk_about_dialog_set_license(GTK_ABOUT_DIALOG(dialog), license);
	gtk_about_dialog_set_website(GTK_ABOUT_DIALOG(dialog), "http://linuxonly.ru");
	gtk_about_dialog_set_website_label(GTK_ABOUT_DIALOG(dialog), "linuxonly.ru");
	
	gtk_about_dialog_set_authors(GTK_ABOUT_DIALOG(dialog), authors);
	gtk_about_dialog_set_documenters(GTK_ABOUT_DIALOG(dialog), documenters);
	gtk_about_dialog_set_artists(GTK_ABOUT_DIALOG(dialog), artists);
	gtk_about_dialog_set_translator_credits(GTK_ABOUT_DIALOG(dialog), _("Alex <alex@linuxonly.ru>"));
	
	gtk_dialog_run(GTK_DIALOG(dialog));
	
	gtk_widget_destroy(dialog);
	
	return 0;
}

static gint cdsdwindow_readhelp(GtkWidget *widget, GdkEvent *event, gpointer data)
{
	GtkWidget *dialog, *content;
	GtkWidget *memo, *sw;
	GtkTextBuffer *buffer;
	GtkTextIter end;
	
	const gchar *help_text = 
		_("Description\n"
		"CDSD is simple program for optical drive speed control.\n"
		"\n"
		"Settings\n"
		"CD speed - set upper limit of read speed for CD (-ROM -R -RW) media. "
		"1x CD speed equals 177 kb/s datarate.\n"
		"DVD speed - set upper limit of read speed for DVD (-ROM -RAM -R -RW +R +RW) media. "
		"1x DVD speed equals 1385 kb/s datarate.\n"
		"BD speed - set upper limit of read speed for BD (-ROM -R -RE) media. "
		"1x BD speed equals 5540 kb/s datarate.\n"
		"Polling - get permission to automatically unmount filesystem when media "
		"ejected with button on front panel of drive.\n"
		"\n"
		"Device info\n"
		"Vendor - device manufacturer identification string.\n"
		"Product - device model identification string.\n"
		"Version - device firmware version.\n"
		"Serial - device serial number.\n"
		"Buffer size - device buffer size in kb. Bigger is better.\n"
		"Maximum read/write speed - device maximum read/write speed for current media type.\n"
		"Current read/write speed - read/write speed set for current media.\n"
		"Media inserted - 1 if media is in drive, 0 if not.\n"
		"Media is eraseable - 1 if drive is write-capable and/or media can be written, 0 if not.\n"
		"Media is CD/DVD/BD - 1 determines type of media inserted.\n"
		"\n"
		"Device properties\n"
		"Table shows media types that drive can read. It is based on device response.");
	
	dialog = gtk_dialog_new_with_buttons(_("Help"), GTK_WINDOW(cdsdwindow.win),
										GTK_DIALOG_MODAL,
										GTK_STOCK_CLOSE, GTK_RESPONSE_CLOSE,
										NULL);
	gtk_dialog_set_default_response(GTK_DIALOG(dialog), GTK_RESPONSE_CLOSE);
	content = gtk_dialog_get_content_area(GTK_DIALOG(dialog));
		
	sw = gtk_scrolled_window_new(NULL, NULL);
	gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(sw), GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);
	gtk_scrolled_window_set_shadow_type(GTK_SCROLLED_WINDOW(sw), GTK_SHADOW_ETCHED_IN);
	gtk_container_set_border_width(GTK_CONTAINER(sw), 5);
	
	memo = gtk_text_view_new();
	
	buffer = gtk_text_view_get_buffer(GTK_TEXT_VIEW(memo));
	gtk_text_buffer_create_tag(buffer, "text", "scale", PANGO_SCALE_MEDIUM, "foreground", "black", NULL);
	
	gtk_text_view_set_editable(GTK_TEXT_VIEW(memo), FALSE);
	gtk_text_view_set_cursor_visible(GTK_TEXT_VIEW(memo), FALSE);
	gtk_text_view_set_wrap_mode(GTK_TEXT_VIEW(memo), GTK_WRAP_WORD_CHAR);
	
	gtk_text_buffer_get_end_iter(GTK_TEXT_BUFFER(buffer), &end);
	gtk_text_buffer_insert_with_tags_by_name(buffer, &end, help_text, -1, "text",  NULL);
	
	gtk_container_add(GTK_CONTAINER(sw), memo);
	gtk_widget_set_size_request(GTK_WIDGET(dialog), 450, 300);
	gtk_container_set_border_width(GTK_CONTAINER(content), 5);
	gtk_box_pack_start(GTK_BOX(content), GTK_WIDGET(sw), TRUE, TRUE, 0);
	gtk_widget_show_all(dialog);
		
	gtk_dialog_run(GTK_DIALOG(dialog));
	
	gtk_widget_destroy(dialog);
	
	return 0;
}

static gboolean cdsdwindow_info_timer_foreach(GtkTreeModel *model, GtkTreePath *path, GtkTreeIter *iter, gpointer data)
{
	struct cdsd_info_timer_data *timerdata;
	guint rowid;
	gchar buffer[32];
	
	timerdata = (struct cdsd_info_timer_data *)data;
	
	gtk_tree_model_get(model, iter, INFO_ID, &rowid, -1);
	
	if (timerdata->infotable[rowid].datatype == INFO_DATA_STRING) {
		g_snprintf((char *)&buffer, sizeof(buffer), timerdata->infotable[rowid].format, (gchar *)(timerdata->infotable[rowid].value));
	} else if (timerdata->infotable[rowid].datatype == INFO_DATA_SHORT) {
		g_snprintf((char *)&buffer, sizeof(buffer), timerdata->infotable[rowid].format, *(gushort *)(timerdata->infotable[rowid].value));
	} else if (timerdata->infotable[rowid].datatype == INFO_DATA_BYTE) {
		g_snprintf((char *)&buffer, sizeof(buffer), timerdata->infotable[rowid].format, *(guchar *)(timerdata->infotable[rowid].value));
	}
	
	gtk_list_store_set(GTK_LIST_STORE(model), iter, INFO_VALUE, buffer, -1);
	
	return FALSE;
}

static gboolean cdsdwindow_info_timer(gpointer data)
{
	struct cdsd_info_timer_data *timerdata;
	gint error;
	
	timerdata = (struct cdsd_info_timer_data *)data;
	
	if (libcdsd_command_info_dl(&error, cdsdwindow.devid, timerdata->info) == 1) {
		gtk_tree_model_foreach(GTK_TREE_MODEL(timerdata->store), cdsdwindow_info_timer_foreach, timerdata);
	}
	
	return TRUE;
}

static gint cdsdwindow_info(GtkWidget *widget, GdkEvent *event, gpointer data)
{
	struct cdsd_settings_info info;
	struct cdsd_info_timer_data timerdata;
	gint timer, error, i;
	GtkWidget *dialog, *content;
	GtkWidget *infolist, *sw;
	GtkTreeSelection *selection;
	GtkCellRenderer *renderer;
	GtkTreeViewColumn *column;
	GtkListStore *store;
	GtkTreeIter iter;
	gchar buffer[32];
	
	cdsd_info_table infotable[] = {
		{_("Vendor"),              "%s",      INFO_DATA_STRING, (void *)(&info.vendorid)},
		{_("Product"),             "%s",      INFO_DATA_STRING, (void *)(&info.productid)},
		{_("Version"),             "%s",      INFO_DATA_STRING, (void *)(&info.productver)},
		{_("Serial"),              "%s",      INFO_DATA_STRING, (void *)(&info.productsn)},
		{_("Buffer size"),         "%u kb",   INFO_DATA_SHORT,  (void *)(&info.buffer_size)},
		{_("Maximum read speed"),  "%u kb/s", INFO_DATA_SHORT,  (void *)(&info.max_read_speed)},
		{_("Current read speed"),  "%u kb/s", INFO_DATA_SHORT,  (void *)(&info.cur_read_speed)},
		{_("Maximum write speed"), "%u kb/s", INFO_DATA_SHORT,  (void *)(&info.max_write_speed)},
		{_("Current write speed"), "%u kb/s", INFO_DATA_SHORT,  (void *)(&info.cur_write_speed)},
		{_("Media inserted"),      "%u",      INFO_DATA_BYTE,   (void *)(&info.media_exists)},
		{_("Media is eraseable"),  "%u",      INFO_DATA_BYTE,   (void *)(&info.media_erasable)},
		{_("Media is CD"),         "%u",      INFO_DATA_BYTE,   (void *)(&info.media_type_cd)},
		{_("Media is DVD"),        "%u",      INFO_DATA_BYTE,   (void *)(&info.media_type_dvd)},
		{_("Media is BD"),         "%u",      INFO_DATA_BYTE,   (void *)(&info.media_type_bd)},
		{NULL,                  NULL,      INFO_DATA_INT,    NULL}
	};
	
		
	if (libcdsd_command_info_dl(&error, cdsdwindow.devid, &info) < 1) {
		cdsdwindow_message(_("Failed to get info for device"), _("Information"), GTK_MESSAGE_INFO, cdsdwindow.win);
	} else {
		dialog = gtk_dialog_new_with_buttons(_("Device info"), GTK_WINDOW(cdsdwindow.win),
											GTK_DIALOG_MODAL,
											GTK_STOCK_CLOSE, GTK_RESPONSE_CLOSE,
											NULL);
		gtk_dialog_set_default_response(GTK_DIALOG(dialog), GTK_RESPONSE_CLOSE);
		content = gtk_dialog_get_content_area(GTK_DIALOG(dialog));
		
		sw = gtk_scrolled_window_new(NULL, NULL);
		gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(sw), GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);
		gtk_container_set_border_width(GTK_CONTAINER(sw), 5);
		
		infolist = gtk_tree_view_new();
		gtk_tree_view_set_headers_visible(GTK_TREE_VIEW(infolist), FALSE);
		gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(sw), GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);
		gtk_scrolled_window_set_shadow_type(GTK_SCROLLED_WINDOW(sw), GTK_SHADOW_ETCHED_IN);
		gtk_tree_view_set_rules_hint(GTK_TREE_VIEW(infolist), TRUE);
		selection = gtk_tree_view_get_selection(GTK_TREE_VIEW(infolist));
		gtk_tree_selection_set_mode(selection, GTK_SELECTION_NONE);
				
		renderer = gtk_cell_renderer_text_new();
		column = gtk_tree_view_column_new_with_attributes("Description", renderer, "text", INFO_DESCRIPTION, NULL);
		gtk_tree_view_append_column(GTK_TREE_VIEW(infolist), column);
		renderer = gtk_cell_renderer_text_new();
		g_object_set(G_OBJECT(renderer), "ellipsize", PANGO_ELLIPSIZE_END, "ellipsize-set", TRUE, NULL);
		column = gtk_tree_view_column_new_with_attributes("Value", renderer, "text", INFO_VALUE, NULL);
		gtk_tree_view_append_column(GTK_TREE_VIEW(infolist), column);
		
		gtk_container_add(GTK_CONTAINER(sw), infolist);
		gtk_widget_set_size_request(GTK_WIDGET(dialog), 450, 300);
		gtk_container_set_border_width(GTK_CONTAINER(content), 5);
		gtk_box_pack_start(GTK_BOX(content), GTK_WIDGET(sw), TRUE, TRUE, 0);
		gtk_widget_show_all(dialog);
		
		i = 0;
		
		store = gtk_list_store_new(INFO_COLUMNS, G_TYPE_STRING, G_TYPE_STRING, G_TYPE_UINT);
		
		while (infotable[i].value != NULL) {
			if (infotable[i].datatype == INFO_DATA_STRING) {
				g_snprintf((char *)&buffer, sizeof(buffer), infotable[i].format, (gchar *)(infotable[i].value));
			} else if (infotable[i].datatype == INFO_DATA_SHORT) {
				g_snprintf((char *)&buffer, sizeof(buffer), infotable[i].format, *(gushort *)(infotable[i].value));
			} else if (infotable[i].datatype == INFO_DATA_BYTE) {
				g_snprintf((char *)&buffer, sizeof(buffer), infotable[i].format, *(guchar *)(infotable[i].value));
			}
			
			gtk_list_store_append(store, &iter);
			gtk_list_store_set(store, &iter, INFO_DESCRIPTION, infotable[i].description, INFO_VALUE, buffer, INFO_ID, i, -1);
			i++;
		}
		
		gtk_tree_view_set_model(GTK_TREE_VIEW(infolist), GTK_TREE_MODEL(store));
		
		timerdata.infotable = infotable;
		timerdata.store = store;
		timerdata.info = &info;
		
		timer = g_timeout_add(3000, cdsdwindow_info_timer, &timerdata);
		
		gtk_dialog_run(GTK_DIALOG(dialog));
		
		g_source_remove(timer);
		
		gtk_widget_destroy(dialog);
	}
	
	return 0;
}

static gint cdsdwindow_capabilities(GtkWidget *widget, GdkEvent *event, gpointer data)
{
	struct cdsd_capabilities_info capabilities;
	gint error, i;
	GtkWidget *dialog, *content;
	GtkWidget *infolist, *sw;
	GtkTreeSelection *selection;
	GtkCellRenderer *renderer;
	GtkTreeViewColumn *column;
	GtkListStore *store;
	GtkTreeIter iter;
	gchar buffer[32];
	
	cdsd_info_table infotable[] = {
		{"AUDIO CD",        "%s",      INFO_DATA_STRING, (void *)(&capabilities.cap_cdrom)},
		{"CD-R",            "%s",      INFO_DATA_STRING, (void *)(&capabilities.cap_cd_r)},
		{"CD-RW",           "%s",      INFO_DATA_STRING, (void *)(&capabilities.cap_cd_rw)},
		{"DVD",             "%s",      INFO_DATA_STRING, (void *)(&capabilities.cap_dvd_rom)},
		{"DVDRAM",          "%s",      INFO_DATA_STRING, (void *)(&capabilities.cap_dvd_ram)},
		{"DVD-R",           "%s",      INFO_DATA_STRING, (void *)(&capabilities.cap_dvd_r)},
		{"DVD-R DL",        "%s",      INFO_DATA_STRING, (void *)(&capabilities.cap_dvd_r_dl)},
		{"DVD-RW",          "%s",      INFO_DATA_STRING, (void *)(&capabilities.cap_dvd_rw)},
		{"DVD-RW DL",       "%s",      INFO_DATA_STRING, (void *)(&capabilities.cap_dvd_rw_dl)},
		{"DVD+R",           "%s",      INFO_DATA_STRING, (void *)(&capabilities.cap_dvd_plus_r)},
		{"DVD+R DL",        "%s",      INFO_DATA_STRING, (void *)(&capabilities.cap_dvd_plus_r_dl)},
		{"DVD+RW",          "%s",      INFO_DATA_STRING, (void *)(&capabilities.cap_dvd_plus_rw)},
		{"DVD+RW DL",       "%s",      INFO_DATA_STRING, (void *)(&capabilities.cap_dvd_plus_rw_dl)},
		{"BD",              "%s",      INFO_DATA_STRING, (void *)(&capabilities.cap_bd_rom)},
		{"BD-R",            "%s",      INFO_DATA_STRING, (void *)(&capabilities.cap_bd_r)},
		{"BD-RE",           "%s",      INFO_DATA_STRING, (void *)(&capabilities.cap_bd_re)},
		{"HDDVD",           "%s",      INFO_DATA_STRING, (void *)(&capabilities.cap_hddvd_rom)},
		{"HDDVD-R",         "%s",      INFO_DATA_STRING, (void *)(&capabilities.cap_hddvd_r)},
		{"HDDVD-R DL",      "%s",      INFO_DATA_STRING, (void *)(&capabilities.cap_hddvd_r_dl)},
		{"HDDVD-RW",        "%s",      INFO_DATA_STRING, (void *)(&capabilities.cap_hddvd_rw)},
		{"HDDVD-RW DL",     "%s",      INFO_DATA_STRING, (void *)(&capabilities.cap_hddvd_rw_dl)},
		{NULL,              NULL,      INFO_DATA_INT,    NULL}
	};
	
		
	if (libcdsd_command_capabilities_dl(&error, cdsdwindow.devid, &capabilities) < 1) {
		cdsdwindow_message(_("Failed to get capabilities info for device"), _("Information"), GTK_MESSAGE_INFO, cdsdwindow.win);
	} else {
		dialog = gtk_dialog_new_with_buttons(_("Device capabilities"), GTK_WINDOW(cdsdwindow.win),
											GTK_DIALOG_MODAL,
											GTK_STOCK_CLOSE, GTK_RESPONSE_CLOSE,
											NULL);
		gtk_dialog_set_default_response(GTK_DIALOG(dialog), GTK_RESPONSE_CLOSE);
		content = gtk_dialog_get_content_area(GTK_DIALOG(dialog));
		
		sw = gtk_scrolled_window_new(NULL, NULL);
		gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(sw), GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);
		gtk_container_set_border_width(GTK_CONTAINER(sw), 5);
		
		infolist = gtk_tree_view_new();
		gtk_tree_view_set_headers_visible(GTK_TREE_VIEW(infolist), FALSE);
		gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(sw), GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);
		gtk_scrolled_window_set_shadow_type(GTK_SCROLLED_WINDOW(sw), GTK_SHADOW_ETCHED_IN);
		gtk_tree_view_set_rules_hint(GTK_TREE_VIEW(infolist), TRUE);
		selection = gtk_tree_view_get_selection(GTK_TREE_VIEW(infolist));
		gtk_tree_selection_set_mode(selection, GTK_SELECTION_NONE);
				
		renderer = gtk_cell_renderer_text_new();
		column = gtk_tree_view_column_new_with_attributes("Description", renderer, "text", INFO_DESCRIPTION, NULL);
		gtk_tree_view_append_column(GTK_TREE_VIEW(infolist), column);
		
		
		renderer = gtk_cell_renderer_pixbuf_new();
		column = gtk_tree_view_column_new_with_attributes("Value", renderer, "stock-id", INFO_VALUE, NULL);
		gtk_tree_view_append_column(GTK_TREE_VIEW(infolist), column);
		
		gtk_container_add(GTK_CONTAINER(sw), infolist);
		gtk_widget_set_size_request(GTK_WIDGET(dialog), 450, 300);
		gtk_container_set_border_width(GTK_CONTAINER(content), 5);
		gtk_box_pack_start(GTK_BOX(content), GTK_WIDGET(sw), TRUE, TRUE, 0);
		gtk_widget_show_all(dialog);
		
		i = 0;
		
		store = gtk_list_store_new(INFO_COLUMNS, G_TYPE_STRING, G_TYPE_STRING, G_TYPE_UINT);
		
		while (infotable[i].value != NULL) {
			if (*(guchar *)(infotable[i].value)) {
				g_snprintf((char *)&buffer, sizeof(buffer), "%s", GTK_STOCK_YES);
			} else {
				g_snprintf((char *)&buffer, sizeof(buffer), "%s", GTK_STOCK_NO);
			}
			
			gtk_list_store_append(store, &iter);
			gtk_list_store_set(store, &iter, INFO_DESCRIPTION, infotable[i].description, INFO_VALUE, buffer, INFO_ID, i, -1);
			i++;
		}
		
		gtk_tree_view_set_model(GTK_TREE_VIEW(infolist), GTK_TREE_MODEL(store));
		
		gtk_dialog_run(GTK_DIALOG(dialog));
		
		gtk_widget_destroy(dialog);
	}
	
	return 0;
}

static gboolean cdsdwindow_wait_dialog_timer(gpointer data)
{
	cdsd_longterm_command *command;
	gint error;
	
	command = (cdsd_longterm_command *)data;
	
	if (libcdsd_command_modeset_dl(&error, command->device) < 1) {
		gtk_dialog_response(GTK_DIALOG(command->dialog), GTK_RESPONSE_NO);
	} else {
		gtk_dialog_response(GTK_DIALOG(command->dialog), GTK_RESPONSE_YES);
	}
		
	return FALSE;
}

static gboolean cdsdwindow_wait_dialog(gpointer data)
{
	cdsd_longterm_command command;
	GtkWidget *dialog, *content;
	GtkWidget *image, *text, *hbox;
	gint timer, response;
	gboolean result;
	
	if (data == NULL) return 0;
	
	gtk_widget_set_sensitive(cdsdwindow.devicesmenu, FALSE);
	
	dialog = gtk_dialog_new_with_buttons(_("Command execution"), GTK_WINDOW(cdsdwindow.win),
										GTK_DIALOG_MODAL,
										NULL);
	gtk_dialog_set_default_response(GTK_DIALOG(dialog), GTK_RESPONSE_OK);
	content = gtk_dialog_get_content_area(GTK_DIALOG(dialog));
	
	hbox = gtk_hbox_new(FALSE, 0);
	
	image = gtk_image_new_from_stock(GTK_STOCK_CDROM, GTK_ICON_SIZE_DIALOG);
	
	text = gtk_label_new(_("Please wait while executing command"));
	
	gtk_box_pack_start(GTK_BOX(hbox), image, TRUE, TRUE, 10);
	gtk_box_pack_start(GTK_BOX(hbox), text, TRUE, TRUE, 10);
	
	gtk_widget_set_size_request(GTK_WIDGET(dialog), 400, 80);
	gtk_container_set_border_width(GTK_CONTAINER(content), 15);
	gtk_box_pack_start(GTK_BOX(content), GTK_WIDGET(hbox), TRUE, TRUE, 0);
	gtk_widget_show_all(dialog);
	
	command.dialog = dialog;
	command.device = (struct cdsd_settings_device *)data;
	
	timer = g_timeout_add(100, cdsdwindow_wait_dialog_timer, &command);
	
	response = gtk_dialog_run(GTK_DIALOG(dialog));
	
	g_source_remove(timer);
	
	gtk_widget_destroy(dialog);
	
	if (response == GTK_RESPONSE_NO) {
		cdsdwindow_message(_("Failed to set speeds for device"), _("Information"), GTK_MESSAGE_INFO, cdsdwindow.win);
		result = FALSE;
	} else {
		result = TRUE;
	}
	
	gtk_widget_set_sensitive(cdsdwindow.devicesmenu, TRUE);
	
	return result;
}

static gint cdsdwindow_apply(GtkWidget *widget, GdkEvent *event, gpointer data)
{
	GtkTreeSelection *selection;
	GtkTreeModel *model;
	GtkTreeIter iter;
	gint setcdindex, setdvdindex, setbdindex, setpolling; 
	struct cdsd_settings_device device;
	gint error;
	
	setcdindex = gtk_combo_box_get_active(GTK_COMBO_BOX(cdsdwindow.cdspeed));
	setdvdindex = gtk_combo_box_get_active(GTK_COMBO_BOX(cdsdwindow.dvdspeed));
	setbdindex = gtk_combo_box_get_active(GTK_COMBO_BOX(cdsdwindow.bdspeed));
	setpolling = gtk_combo_box_get_active(GTK_COMBO_BOX(cdsdwindow.polling));
	
	device.id = cdsdwindow.devid;
	device.cdspeed = cdspeeds[setcdindex];
	device.dvdspeed = dvdspeeds[setdvdindex];
	device.bdspeed = bdspeeds[setbdindex];
	device.polling = setpolling;
	
	if (cdsdwindow_wait_dialog(&device)) {
		selection = gtk_tree_view_get_selection(GTK_TREE_VIEW(cdsdwindow.list));
		if (gtk_tree_selection_get_selected(selection, &model, &iter)) {
			gtk_list_store_set(GTK_LIST_STORE(model), &iter, DEVS_COL_CDSPEED, device.cdspeed, DEVS_COL_DVDSPEED, device.dvdspeed, DEVS_COL_BDSPEED, device.bdspeed, DEVS_COL_POLLING, device.polling, -1);
		}
		
		cdsdwindow.cdspeedindex = setcdindex;
		cdsdwindow.dvdspeedindex = setdvdindex;
		cdsdwindow.bdspeedindex = setbdindex;
		cdsdwindow.pollingindex = setpolling;
	}
	
	cdsdwindow_set_action_buttons_state(TRUE, FALSE);
	
	return 0;
}

static void cdsdwindow_change_parameter(GtkComboBox *widget, gpointer data)
{
	gint i;
	gint setcdindex, setdvdindex, setbdindex, setpolling; 
	gboolean newdata;
	
	typedef struct {
		gint *defvalue;
		gint *newvalue;
	} values_t;
	
	values_t values[] = {
		{(gint *)(&cdsdwindow.cdspeedindex), (gint *)(&setcdindex)},
		{(gint *)(&cdsdwindow.dvdspeedindex), (gint *)(&setdvdindex)},
		{(gint *)(&cdsdwindow.bdspeedindex), (gint *)(&setbdindex)},
		{(gint *)(&cdsdwindow.pollingindex), (gint *)(&setpolling)},
		{NULL, NULL}
	};
	
	setcdindex = gtk_combo_box_get_active(GTK_COMBO_BOX(cdsdwindow.cdspeed));
	setdvdindex = gtk_combo_box_get_active(GTK_COMBO_BOX(cdsdwindow.dvdspeed));
	setbdindex = gtk_combo_box_get_active(GTK_COMBO_BOX(cdsdwindow.bdspeed));
	setpolling = gtk_combo_box_get_active(GTK_COMBO_BOX(cdsdwindow.polling));
	
	i = 0;
	newdata = FALSE;
	while (values[i].defvalue != NULL) {
		if (*(values[i].defvalue) != *(values[i].newvalue)) {
			newdata = TRUE;
			break;
		}
		i++;
	}
	
	cdsdwindow_set_action_buttons_state(TRUE, newdata);
}

static void cdsdwindow_init_list(GtkWidget *list)
{
	GtkCellRenderer    *renderer;
	GtkTreeViewColumn  *column1, *column2;
	GtkListStore       *store;
		
	
	renderer = gtk_cell_renderer_pixbuf_new();
	column1 = gtk_tree_view_column_new();
	gtk_tree_view_column_set_title(column1, NULL);
	gtk_tree_view_column_pack_start(column1, renderer, FALSE);
	gtk_tree_view_column_set_attributes(column1, renderer, "stock-id", DEVS_COL_ICON, "stock-size", DEVS_COL_ICON_SIZE, NULL);
	gtk_tree_view_append_column(GTK_TREE_VIEW(list), column1);
	
	renderer = gtk_cell_renderer_text_new();
	g_object_set(G_OBJECT(renderer), "ellipsize", PANGO_ELLIPSIZE_END, "ellipsize-set", TRUE, NULL);
	column2 = gtk_tree_view_column_new_with_attributes(_("Device"), renderer, "markup", DEVS_COL_NAME, NULL);
	gtk_tree_view_append_column(GTK_TREE_VIEW(list), column2);
	
	store = gtk_list_store_new(DEVS_COLUMNS, G_TYPE_STRING, G_TYPE_INT, G_TYPE_STRING, G_TYPE_INT, G_TYPE_INT, G_TYPE_INT, G_TYPE_INT, G_TYPE_INT);
	
	gtk_tree_view_set_model(GTK_TREE_VIEW(list), GTK_TREE_MODEL(store));
	g_object_unref(store);
}

static gboolean cdsdwindow_set_controls(gint control, gint value, gint instance)
{
	gint i;
	gboolean found;
	
	i = 0;
	found = FALSE;
	
	switch (control) {
		case CONTROLS_CD_SPEED: {
			while (cdspeeds[i] != -1) {
				if (value == cdspeeds[i]) {
					found = TRUE;
					break;
				}
				i++;
			}
			if (!found) i = 0; 
			if (instance == INSTANCE_WINDOW) {
				gtk_combo_box_set_active(GTK_COMBO_BOX(cdsdwindow.cdspeed), i);
				cdsdwindow.cdspeedindex = i;
			} else if (instance == INSTANCE_TRAY) {
				gtk_check_menu_item_set_active(GTK_CHECK_MENU_ITEM(cdsdwindow.cdspeed_tm[i]), TRUE);
			}
			break;
		}
		
		case CONTROLS_DVD_SPEED: {
			while (dvdspeeds[i] != -1) {
				if (value == dvdspeeds[i]) {
					found = TRUE;
					break;
				}
				i++;
			}
			if (!found) i = 0;
			if (instance == INSTANCE_WINDOW) {
				gtk_combo_box_set_active(GTK_COMBO_BOX(cdsdwindow.dvdspeed), i);
				cdsdwindow.dvdspeedindex = i;
			} else if (instance == INSTANCE_TRAY) {
				gtk_check_menu_item_set_active(GTK_CHECK_MENU_ITEM(cdsdwindow.dvdspeed_tm[i]), TRUE);
			}	
			break;
		}
		
		case CONTROLS_BD_SPEED: {
			while (bdspeeds[i] != -1) {
				if (value == bdspeeds[i]) {
					found = TRUE;
					break;
				}
				i++;
			}
			if (!found) i = 0; 
			if (instance == INSTANCE_WINDOW) {
				gtk_combo_box_set_active(GTK_COMBO_BOX(cdsdwindow.bdspeed), i);
				cdsdwindow.bdspeedindex = i;
			} else if (instance == INSTANCE_TRAY) {
				gtk_check_menu_item_set_active(GTK_CHECK_MENU_ITEM(cdsdwindow.bdspeed_tm[i]), TRUE);
			}
			break;
		}
		
		case CONTROLS_POLLING: {
			if (value == 1) i = 1; else i = 0;
			found = TRUE;
			gtk_combo_box_set_active(GTK_COMBO_BOX(cdsdwindow.polling), i);
			cdsdwindow.pollingindex = i;
			break;
		}
		
		default: {
			return FALSE;
		}
	}
	
	return found;
}

static gboolean cdsdwindow_device_set_speed_tray_foreach(GtkTreeModel *model, GtkTreePath *path, GtkTreeIter *iter, gpointer data)
{
	gint cmd, parameter, speed;
	gint id, cdspeed, dvdspeed, bdspeed, polling;
	gboolean valchanged;
	struct cdsd_settings_device device;
	
	gtk_tree_model_get(model, iter, DEVS_COL_ID, &id, DEVS_COL_CDSPEED, &cdspeed, DEVS_COL_DVDSPEED, &dvdspeed, DEVS_COL_BDSPEED, &bdspeed, DEVS_COL_POLLING, &polling, -1);
	
	if (cdsdwindow.traydevid == id) {
		cmd = GPOINTER_TO_INT(data);
		parameter = CDSD_SPEED_TRAY_UNPACK1(cmd);
		speed = CDSD_SPEED_TRAY_UNPACK2(cmd);
		
		valchanged = FALSE;
		
		device.id = cdsdwindow.traydevid;
		device.cdspeed = cdspeed;
		device.dvdspeed = dvdspeed;
		device.bdspeed = bdspeed;
		device.polling = polling;
		
		switch (parameter) {
			case TRAY_CDSPEED:
				if (cdspeeds[speed] != cdspeed) {
					device.cdspeed = cdspeeds[speed];
					valchanged = TRUE;
				}
				break;
			case TRAY_DVDSPEED:
				if (dvdspeeds[speed] != dvdspeed) {
					device.dvdspeed = dvdspeeds[speed];
					valchanged = TRUE;
				}
				break;
			case TRAY_BDSPEED:
				if (bdspeeds[speed] != bdspeed) {
					device.bdspeed = bdspeeds[speed];
					valchanged = TRUE;
				}
				break;
			default:
				valchanged = FALSE;
				break;
		}
		
		if (valchanged) {
			if (cdsdwindow_wait_dialog(&device)) {
				gtk_list_store_set(GTK_LIST_STORE(model), iter, DEVS_COL_CDSPEED, device.cdspeed, DEVS_COL_DVDSPEED, device.dvdspeed, DEVS_COL_BDSPEED, device.bdspeed, DEVS_COL_POLLING, device.polling, -1);
				
				if ((gtk_tree_view_get_selection(GTK_TREE_VIEW(cdsdwindow.list)) != NULL) && (cdsdwindow.traydevid == cdsdwindow.devid)) {
					cdsdwindow_set_controls(CONTROLS_CD_SPEED, device.cdspeed, INSTANCE_WINDOW);
					cdsdwindow_set_controls(CONTROLS_DVD_SPEED, device.dvdspeed, INSTANCE_WINDOW);
					cdsdwindow_set_controls(CONTROLS_BD_SPEED, device.bdspeed, INSTANCE_WINDOW);
					cdsdwindow_set_controls(CONTROLS_POLLING, device.polling, INSTANCE_WINDOW);
					
					cdsdwindow_set_action_buttons_state(TRUE, FALSE);
				}
			}
		}
		
		return TRUE;
	}
	
	return FALSE;
}

static void cdsdwindow_device_set_speed_tray(GtkMenuItem *menuitem, gpointer data)
{
	GtkTreeModel *model;
	
	if (!cdsdwindow.traylock) {
		if ((model = gtk_tree_view_get_model(GTK_TREE_VIEW(cdsdwindow.list))) != NULL) {
			gtk_tree_model_foreach(model, cdsdwindow_device_set_speed_tray_foreach, data);
		}
	}
	
}

static gboolean cdsdwindow_device_selected_tray_foreach(GtkTreeModel *model, GtkTreePath *path, GtkTreeIter *iter, gpointer data)
{
	gint devid;
	gint id, cdspeed, dvdspeed, bdspeed;
	
	devid = GPOINTER_TO_INT(data);
	
	gtk_tree_model_get(model, iter, DEVS_COL_ID, &id, DEVS_COL_CDSPEED, &cdspeed, DEVS_COL_DVDSPEED, &dvdspeed, DEVS_COL_BDSPEED, &bdspeed, -1);
	
	if (devid == id) {
		cdsdwindow.traylock = TRUE;
		cdsdwindow_set_controls(CONTROLS_CD_SPEED, cdspeed, INSTANCE_TRAY);
		cdsdwindow_set_controls(CONTROLS_DVD_SPEED, dvdspeed, INSTANCE_TRAY);
		cdsdwindow_set_controls(CONTROLS_BD_SPEED, bdspeed, INSTANCE_TRAY);
		cdsdwindow.traylock = FALSE;
		
		return TRUE;
	}
	
	return FALSE;
}

static void cdsdwindow_device_selected_tray(GtkMenuItem *menuitem, gpointer data)
{
	GtkTreeModel *model;
	
	cdsdwindow.traydevid = GPOINTER_TO_INT(data);
	
	if ((model = gtk_tree_view_get_model(GTK_TREE_VIEW(cdsdwindow.list))) != NULL) {
		gtk_tree_model_foreach(model, cdsdwindow_device_selected_tray_foreach, data);
	}
}

static int cdsdwindow_add_event_cb(struct cdsd_settings_device *device)
{
	GtkTreeModel *model;
	GtkTreeIter iter;
	GtkWidget *menuitem;
	gchar buffer[256];
	
	snprintf(buffer, sizeof(buffer), "<small><b>%s %s</b>\n%s</small>", device->vendor, device->product, device->fname);
	
	if ((model = gtk_tree_view_get_model(GTK_TREE_VIEW(cdsdwindow.list))) != NULL) {
		gtk_list_store_append(GTK_LIST_STORE(model), &iter);
		gtk_list_store_set(GTK_LIST_STORE(model), &iter, DEVS_COL_ICON, GTK_STOCK_CDROM, DEVS_COL_ICON_SIZE, GTK_ICON_SIZE_BUTTON,
							DEVS_COL_NAME, buffer, DEVS_COL_ID, device->id, DEVS_COL_CDSPEED, device->cdspeed, DEVS_COL_DVDSPEED,
							device->dvdspeed, DEVS_COL_BDSPEED, device->bdspeed, DEVS_COL_POLLING, device->polling, -1);
	
	
	}
	
	menuitem = gtk_menu_item_new_with_label(device->product);
	gtk_menu_shell_append(GTK_MENU_SHELL(cdsdwindow.devicesmenu), menuitem);
	gtk_menu_item_set_submenu(GTK_MENU_ITEM(menuitem), cdsdwindow.speedsmenu);
	
	g_signal_connect(G_OBJECT(menuitem), "activate", G_CALLBACK(cdsdwindow_device_selected_tray), GINT_TO_POINTER(device->id));
	
	gtk_widget_show_all(menuitem);
	
	return FALSE;
}


static void cdsdwindow_device_selected(GtkTreeView *tree_view, gpointer data)
{
	GtkTreeSelection *selection;
	GtkTreeModel *model;
	GtkTreeIter iter;
	gint id, cdspeed, dvdspeed, bdspeed, polling;
	
	selection = gtk_tree_view_get_selection(GTK_TREE_VIEW(tree_view));
	if (gtk_tree_selection_get_selected(selection, &model, &iter)) {
		gtk_tree_model_get(model, &iter, DEVS_COL_ID, &id,DEVS_COL_CDSPEED, &cdspeed, DEVS_COL_DVDSPEED,
							&dvdspeed, DEVS_COL_BDSPEED, &bdspeed, DEVS_COL_POLLING, &polling, -1);
		
		cdsdwindow_set_controls(CONTROLS_CD_SPEED, cdspeed, INSTANCE_WINDOW);
		cdsdwindow_set_controls(CONTROLS_DVD_SPEED, dvdspeed, INSTANCE_WINDOW);
		cdsdwindow_set_controls(CONTROLS_BD_SPEED, bdspeed, INSTANCE_WINDOW);
		cdsdwindow_set_controls(CONTROLS_POLLING, polling, INSTANCE_WINDOW);
		
		cdsdwindow.devid = id;
		
		cdsdwindow_set_action_buttons_state(TRUE, FALSE);
	}
}	

static void cdsdwindow_set_action_buttons_state(gboolean enabled, gboolean changed)
{
	gtk_widget_set_sensitive(cdsdwindow.table, enabled);
	gtk_widget_set_sensitive(cdsdwindow.apply, (enabled && changed));
	gtk_widget_set_sensitive(cdsdwindow.apply_menu, (enabled && changed));
	gtk_widget_set_sensitive(cdsdwindow.info_menu, enabled);
	gtk_widget_set_sensitive(cdsdwindow.capabilities_menu, enabled);
}

static gint cdsdwindow_show(GtkWidget *widget, GdkEvent *event, gpointer data)
{
	cdsdwindow_trayaction(GTK_STATUS_ICON(cdsdwindow.trayicon), data);
	
	return 0;
}

static void cdsdwindow_trayaction(GtkStatusIcon *status_icon, gpointer data)
{
	#if GTK_CHECK_VERSION(2, 20, 0)
		if (gtk_widget_get_visible(cdsdwindow.win)) {
	#else
		if (GTK_WIDGET_VISIBLE(cdsdwindow.win)) {
	#endif
		#if GTK_CHECK_VERSION(2, 24, 0)
			gtk_widget_hide(cdsdwindow.win);
		#else
			gtk_widget_hide_all(cdsdwindow.win);
		#endif
		g_signal_handler_block(G_OBJECT(cdsdwindow.showwin_tm), cdsdwindow.traysigid);
		gtk_check_menu_item_set_active(GTK_CHECK_MENU_ITEM(cdsdwindow.showwin_tm), FALSE);
		g_signal_handler_unblock(G_OBJECT(cdsdwindow.showwin_tm), cdsdwindow.traysigid);
	} else {
		gtk_widget_show_all(cdsdwindow.win);
		g_signal_handler_block(G_OBJECT(cdsdwindow.showwin_tm), cdsdwindow.traysigid);
		gtk_check_menu_item_set_active(GTK_CHECK_MENU_ITEM(cdsdwindow.showwin_tm), TRUE);
		g_signal_handler_unblock(G_OBJECT(cdsdwindow.showwin_tm), cdsdwindow.traysigid);
	}
}

static void cdsdwindow_traypopup(GtkStatusIcon *status_icon, guint button, guint activate_time, gpointer data)
{
	gtk_menu_popup(GTK_MENU(data), NULL, NULL, gtk_status_icon_position_menu, status_icon, button, activate_time);
}

static GtkWidget *cdsdwindow_create(void)
{
	GtkWidget *vbox, *hbox;
	GtkWidget *paned, *frame;
	GtkWidget *align;
	GtkWidget *sw;
	
	GtkWidget *caption_l;
	GtkWidget *cdspeed_l;
	GtkWidget *dvdspeed_l;
	GtkWidget *bdspeed_l;
	GtkWidget *polling_l;
	
	gint i;
	gchar textbuf[256];
	GSList *group = NULL;
	
	
	if (!cdsdwindow.visible) {
		cdsdwindow.visible = TRUE;
		
		cdsdwindow.devid = -1;
		cdsdwindow.traydevid = -1;
		
		cdsdwindow.win = gtk_window_new(GTK_WINDOW_TOPLEVEL);
		cdsdwindow.title = _("CDSD control");
		gtk_window_set_position(GTK_WINDOW(cdsdwindow.win), GTK_WIN_POS_CENTER);
		gtk_window_set_title(GTK_WINDOW(cdsdwindow.win), cdsdwindow.title);
		gtk_widget_set_size_request(cdsdwindow.win, 550, 350);
		gtk_window_set_default_size(GTK_WINDOW(cdsdwindow.win), 550, 350);
		gtk_window_set_resizable(GTK_WINDOW(cdsdwindow.win), FALSE);
		
		cdsdwindow.menubar = gtk_menu_bar_new();
		
		cdsdwindow.accel_group = gtk_accel_group_new();
		gtk_window_add_accel_group(GTK_WINDOW(cdsdwindow.win), cdsdwindow.accel_group);
		
		cdsdwindow.devmenu = gtk_menu_new();
		cdsdwindow.dev_menu = gtk_menu_item_new_with_mnemonic(_("_Device"));
		cdsdwindow.apply_menu = gtk_image_menu_item_new_from_stock(GTK_STOCK_APPLY, cdsdwindow.accel_group);
		cdsdwindow.info_menu = gtk_image_menu_item_new_from_stock(GTK_STOCK_INFO, cdsdwindow.accel_group);
		cdsdwindow.capabilities_menu = gtk_image_menu_item_new_from_stock(GTK_STOCK_PROPERTIES, cdsdwindow.accel_group);
		cdsdwindow.sep_menu = gtk_separator_menu_item_new();
		cdsdwindow.quit_menu = gtk_image_menu_item_new_from_stock(GTK_STOCK_QUIT, cdsdwindow.accel_group);
		
		cdsdwindow.helpmenu = gtk_menu_new();
		cdsdwindow.help_menu = gtk_menu_item_new_with_mnemonic(_("_Help"));
		cdsdwindow.readhelp_menu = gtk_image_menu_item_new_from_stock(GTK_STOCK_HELP, cdsdwindow.accel_group);
		cdsdwindow.about_menu = gtk_image_menu_item_new_from_stock(GTK_STOCK_ABOUT, cdsdwindow.accel_group);
		
		gtk_menu_item_set_submenu(GTK_MENU_ITEM(cdsdwindow.dev_menu), cdsdwindow.devmenu);
		gtk_menu_shell_append(GTK_MENU_SHELL(cdsdwindow.devmenu), cdsdwindow.apply_menu);
		gtk_menu_shell_append(GTK_MENU_SHELL(cdsdwindow.devmenu), cdsdwindow.info_menu);
		gtk_menu_shell_append(GTK_MENU_SHELL(cdsdwindow.devmenu), cdsdwindow.capabilities_menu);
		gtk_menu_shell_append(GTK_MENU_SHELL(cdsdwindow.devmenu), cdsdwindow.sep_menu);
		gtk_menu_shell_append(GTK_MENU_SHELL(cdsdwindow.devmenu), cdsdwindow.quit_menu);
		gtk_menu_shell_append(GTK_MENU_SHELL(cdsdwindow.menubar), cdsdwindow.dev_menu);
		
		gtk_menu_item_set_submenu(GTK_MENU_ITEM(cdsdwindow.help_menu), cdsdwindow.helpmenu);
		gtk_menu_shell_append(GTK_MENU_SHELL(cdsdwindow.helpmenu), cdsdwindow.readhelp_menu);
		gtk_menu_shell_append(GTK_MENU_SHELL(cdsdwindow.helpmenu), cdsdwindow.about_menu);
		gtk_menu_shell_append(GTK_MENU_SHELL(cdsdwindow.menubar), cdsdwindow.help_menu);
		
		g_signal_connect(G_OBJECT(cdsdwindow.apply_menu), "activate", G_CALLBACK(cdsdwindow_apply), NULL);
		g_signal_connect(G_OBJECT(cdsdwindow.info_menu), "activate", G_CALLBACK(cdsdwindow_info), NULL);
		g_signal_connect(G_OBJECT(cdsdwindow.capabilities_menu), "activate", G_CALLBACK(cdsdwindow_capabilities), NULL);
		g_signal_connect(G_OBJECT(cdsdwindow.quit_menu), "activate", G_CALLBACK(cdsdwindow_close), NULL);
		
		g_signal_connect(G_OBJECT(cdsdwindow.readhelp_menu), "activate", G_CALLBACK(cdsdwindow_readhelp), NULL);
		g_signal_connect(G_OBJECT(cdsdwindow.about_menu), "activate", G_CALLBACK(cdsdwindow_about), NULL);
				
		vbox = gtk_vbox_new(FALSE, 0);
		hbox = gtk_hbox_new(FALSE, 0);
		
		gtk_box_pack_start(GTK_BOX(vbox), cdsdwindow.menubar, FALSE, FALSE, 0);
		
		sw = gtk_scrolled_window_new(NULL, NULL);
		gtk_container_set_border_width(GTK_CONTAINER(sw), 0);
		
		cdsdwindow.list = gtk_tree_view_new();
		gtk_tree_view_set_headers_visible(GTK_TREE_VIEW(cdsdwindow.list), FALSE);
		gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(sw), GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);
		gtk_scrolled_window_set_shadow_type(GTK_SCROLLED_WINDOW(sw), GTK_SHADOW_ETCHED_IN);
		
		gtk_container_add(GTK_CONTAINER (sw), cdsdwindow.list);
		cdsdwindow.selection = gtk_tree_view_get_selection(GTK_TREE_VIEW(cdsdwindow.list));
		gtk_tree_selection_set_mode(cdsdwindow.selection, GTK_SELECTION_BROWSE);
		
		cdsdwindow_init_list(cdsdwindow.list);
		
		gtk_tree_view_set_rules_hint(GTK_TREE_VIEW(cdsdwindow.list), TRUE);
		
		//Add panel there
		cdsdwindow.table = gtk_table_new(2, 5, FALSE);
		
		caption_l = gtk_label_new(NULL);
		gtk_label_set_markup(GTK_LABEL(caption_l), _("<b>Device settings</b>"));
		gtk_label_set_max_width_chars(GTK_LABEL(caption_l), 30);
		gtk_label_set_ellipsize(GTK_LABEL(caption_l), PANGO_ELLIPSIZE_END);
		gtk_table_attach(GTK_TABLE(cdsdwindow.table), caption_l, 0, 2, 0, 1, GTK_SHRINK, GTK_SHRINK, 0, 10);
		
		cdspeed_l = gtk_label_new(_("CD speed"));
		align = gtk_alignment_new(1.0, 0.5, 0.0, 0.0);
		gtk_container_add(GTK_CONTAINER(align), cdspeed_l);
		gtk_table_attach(GTK_TABLE(cdsdwindow.table), align, 0, 1, 1, 2, GTK_FILL, GTK_SHRINK, 5, 5);
				
		#if GTK_CHECK_VERSION(2, 24, 0)
			cdsdwindow.cdspeed = gtk_combo_box_text_new();
			i = 0;
			while (cdspeeds[i] != -1) {
				if (cdspeeds[i] == 0) {
					gtk_combo_box_text_append_text(GTK_COMBO_BOX_TEXT(cdsdwindow.cdspeed), _("Optimal"));
				} else {
					snprintf((gchar *)&textbuf, sizeof(textbuf), "%u x", cdspeeds[i]);
					gtk_combo_box_text_append_text(GTK_COMBO_BOX_TEXT(cdsdwindow.cdspeed), (gchar *)&textbuf);
				}
				i++;
			}
		#else
			cdsdwindow.cdspeed = gtk_combo_box_new_text();
			i = 0;
			while (cdspeeds[i] != -1) {
				if (cdspeeds[i] == 0) {
					gtk_combo_box_append_text(GTK_COMBO_BOX(cdsdwindow.cdspeed), _("Optimal"));
				} else {
					snprintf((gchar *)&textbuf, sizeof(textbuf), "%u x", cdspeeds[i]);
					gtk_combo_box_append_text(GTK_COMBO_BOX(cdsdwindow.cdspeed), (gchar *)&textbuf);
				}
				i++;
			}
		#endif
		
		gtk_combo_box_set_active(GTK_COMBO_BOX(cdsdwindow.cdspeed), 0);
		g_signal_connect(G_OBJECT(cdsdwindow.cdspeed), "changed", G_CALLBACK(cdsdwindow_change_parameter), GINT_TO_POINTER(CONTROLS_CD_SPEED));
		gtk_table_attach(GTK_TABLE(cdsdwindow.table), cdsdwindow.cdspeed, 1, 2, 1, 2, GTK_EXPAND | GTK_FILL, GTK_SHRINK, 5, 5);
		
		dvdspeed_l = gtk_label_new(_("DVD speed"));
		align = gtk_alignment_new(1.0, 0.5, 0.0, 0.0);
		gtk_container_add(GTK_CONTAINER(align), dvdspeed_l);
		gtk_table_attach(GTK_TABLE(cdsdwindow.table), align, 0, 1, 2, 3, GTK_FILL, GTK_SHRINK, 5, 5);
		
		#if GTK_CHECK_VERSION(2, 24, 0)
			cdsdwindow.dvdspeed = gtk_combo_box_text_new();
			i = 0;
			while (dvdspeeds[i] != -1) {
				if (dvdspeeds[i] == 0) {
					gtk_combo_box_text_append_text(GTK_COMBO_BOX_TEXT(cdsdwindow.dvdspeed), _("Optimal"));
				} else {
					snprintf((gchar *)&textbuf, sizeof(textbuf), "%u x", dvdspeeds[i]);
					gtk_combo_box_text_append_text(GTK_COMBO_BOX_TEXT(cdsdwindow.dvdspeed), (gchar *)&textbuf);
				}
				i++;
			}
		#else
			cdsdwindow.dvdspeed = gtk_combo_box_new_text();
			i = 0;
			while (dvdspeeds[i] != -1) {
				if (dvdspeeds[i] == 0) {
					gtk_combo_box_append_text(GTK_COMBO_BOX(cdsdwindow.dvdspeed), _("Optimal"));
				} else {
					snprintf((gchar *)&textbuf, sizeof(textbuf), "%u x", dvdspeeds[i]);
					gtk_combo_box_append_text(GTK_COMBO_BOX(cdsdwindow.dvdspeed), (gchar *)&textbuf);
				}
				i++;
			}
		#endif
		
		gtk_combo_box_set_active(GTK_COMBO_BOX(cdsdwindow.dvdspeed), 0);
		g_signal_connect(G_OBJECT(cdsdwindow.dvdspeed), "changed", G_CALLBACK(cdsdwindow_change_parameter), GINT_TO_POINTER(CONTROLS_DVD_SPEED));
		gtk_table_attach(GTK_TABLE(cdsdwindow.table), cdsdwindow.dvdspeed, 1, 2, 2, 3, GTK_EXPAND | GTK_FILL, GTK_SHRINK, 5, 5);
		
		bdspeed_l = gtk_label_new(_("BD speed"));
		align = gtk_alignment_new(1.0, 0.5, 0.0, 0.0);
		gtk_container_add(GTK_CONTAINER(align), bdspeed_l);
		gtk_table_attach(GTK_TABLE(cdsdwindow.table), align, 0, 1, 3, 4, GTK_FILL, GTK_SHRINK, 5, 5);
		
		#if GTK_CHECK_VERSION(2, 24, 0)
			cdsdwindow.bdspeed = gtk_combo_box_text_new();
			i = 0;
			while (bdspeeds[i] != -1) {
				if (bdspeeds[i] == 0) {
					gtk_combo_box_text_append_text(GTK_COMBO_BOX_TEXT(cdsdwindow.bdspeed), _("Optimal"));
				} else {
					snprintf((gchar *)&textbuf, sizeof(textbuf), "%u x", bdspeeds[i]);
					gtk_combo_box_text_append_text(GTK_COMBO_BOX_TEXT(cdsdwindow.bdspeed), (gchar *)&textbuf);
				}
				i++;
			}
		#else	
			cdsdwindow.bdspeed = gtk_combo_box_new_text();
			i = 0;
			while (bdspeeds[i] != -1) {
				if (bdspeeds[i] == 0) {
					gtk_combo_box_append_text(GTK_COMBO_BOX(cdsdwindow.bdspeed), _("Optimal"));
				} else {
					snprintf((gchar *)&textbuf, sizeof(textbuf), "%u x", bdspeeds[i]);
					gtk_combo_box_append_text(GTK_COMBO_BOX(cdsdwindow.bdspeed), (gchar *)&textbuf);
				}
				i++;
			}
		#endif
		
		gtk_combo_box_set_active(GTK_COMBO_BOX(cdsdwindow.bdspeed), 0);
		g_signal_connect(G_OBJECT(cdsdwindow.bdspeed), "changed", G_CALLBACK(cdsdwindow_change_parameter), GINT_TO_POINTER(CONTROLS_BD_SPEED));
		gtk_table_attach(GTK_TABLE(cdsdwindow.table), cdsdwindow.bdspeed, 1, 2, 3, 4, GTK_EXPAND | GTK_FILL, GTK_SHRINK, 5, 5);
		
		polling_l = gtk_label_new(_("Polling"));
		align = gtk_alignment_new(1.0, 0.5, 0.0, 0.0);
		gtk_container_add(GTK_CONTAINER(align), polling_l);
		gtk_table_attach(GTK_TABLE(cdsdwindow.table), align, 0, 1, 4, 5, GTK_FILL, GTK_SHRINK, 5, 5);
		
		#if GTK_CHECK_VERSION(2, 24, 0)
			cdsdwindow.polling = gtk_combo_box_text_new();
			gtk_combo_box_text_append_text(GTK_COMBO_BOX_TEXT(cdsdwindow.polling), _("Off"));
			gtk_combo_box_text_append_text(GTK_COMBO_BOX_TEXT(cdsdwindow.polling), _("On"));
		#else	
			cdsdwindow.polling = gtk_combo_box_new_text();
			gtk_combo_box_append_text(GTK_COMBO_BOX(cdsdwindow.polling), _("Off"));
			gtk_combo_box_append_text(GTK_COMBO_BOX(cdsdwindow.polling), _("On"));
		#endif
		
		gtk_combo_box_set_active(GTK_COMBO_BOX(cdsdwindow.polling), 0);
		g_signal_connect(G_OBJECT(cdsdwindow.polling), "changed", G_CALLBACK(cdsdwindow_change_parameter), GINT_TO_POINTER(CONTROLS_POLLING));
		gtk_table_attach(GTK_TABLE(cdsdwindow.table), cdsdwindow.polling, 1, 2, 4, 5, GTK_EXPAND | GTK_FILL, GTK_SHRINK, 5, 5);
		
		align = gtk_alignment_new(0.5, 0.5, 0.0, 0.0);
		gtk_container_add(GTK_CONTAINER(align), cdsdwindow.table);
		
		#if GTK_CHECK_VERSION(3, 0, 0)
			paned = gtk_paned_new(GTK_ORIENTATION_HORIZONTAL);
		#else
			paned = gtk_hpaned_new();
		#endif
		gtk_container_set_border_width(GTK_CONTAINER(paned), 5);
		
		gtk_paned_pack1(GTK_PANED(paned), sw, TRUE, FALSE);
		gtk_widget_set_size_request(GTK_WIDGET(sw), 150, -1);
		
		gtk_paned_pack2(GTK_PANED(paned), align, TRUE, FALSE);
		
		gtk_box_pack_start(GTK_BOX(vbox), paned, TRUE, TRUE, 0);
		
		cdsdwindow.close = gtk_button_new_from_stock(GTK_STOCK_QUIT);
		cdsdwindow.apply = gtk_button_new_from_stock(GTK_STOCK_APPLY);
				
		cdsdwindow_set_action_buttons_state(FALSE, FALSE);
		
		hbox = gtk_hbox_new(FALSE, 0);
		gtk_box_pack_end(GTK_BOX(hbox), cdsdwindow.close, FALSE, FALSE, 5);
		gtk_box_pack_end(GTK_BOX(hbox), cdsdwindow.apply, FALSE, FALSE, 5);
		
		gtk_box_pack_start(GTK_BOX(vbox), hbox, FALSE, FALSE, 5);
		
		g_signal_connect(G_OBJECT(cdsdwindow.win), "destroy", G_CALLBACK(cdsdwindow_close), NULL);
		g_signal_connect(G_OBJECT(cdsdwindow.close), "clicked", G_CALLBACK(cdsdwindow_close), NULL);
		g_signal_connect(G_OBJECT(cdsdwindow.apply), "clicked", G_CALLBACK(cdsdwindow_apply), NULL);
		g_signal_connect(G_OBJECT(cdsdwindow.list), "cursor-changed", G_CALLBACK(cdsdwindow_device_selected), NULL);
				
		gtk_container_add(GTK_CONTAINER(cdsdwindow.win),vbox);
		gtk_widget_show_all(cdsdwindow.win);
		
		gtk_window_set_focus(GTK_WINDOW(cdsdwindow.win), cdsdwindow.list);
		
		cdsdwindow.trayicon = gtk_status_icon_new_from_file(RESOURCE_MAIN_ICON);
				
		cdsdwindow.traymenu = gtk_menu_new();
		cdsdwindow.showwin_tm =	gtk_check_menu_item_new_with_label(_("Show window"));
		
		#if GTK_CHECK_VERSION(2, 20, 0)
			gtk_check_menu_item_set_active(GTK_CHECK_MENU_ITEM(cdsdwindow.showwin_tm), gtk_widget_get_visible(cdsdwindow.win));
		#else
			gtk_check_menu_item_set_active(GTK_CHECK_MENU_ITEM(cdsdwindow.showwin_tm), GTK_WIDGET_VISIBLE(cdsdwindow.win));
		#endif
		
		cdsdwindow.sep1_tm = gtk_separator_menu_item_new();
		
		cdsdwindow.devices_tm = gtk_image_menu_item_new_with_label(_("Devices"));
		
		cdsdwindow.sep2_tm = gtk_separator_menu_item_new();
		cdsdwindow.quit_tm = gtk_image_menu_item_new_from_stock(GTK_STOCK_QUIT, NULL);
		
		cdsdwindow.traysigid = g_signal_connect(G_OBJECT(cdsdwindow.showwin_tm), "activate", G_CALLBACK(cdsdwindow_show), NULL);
		g_signal_connect(G_OBJECT(cdsdwindow.quit_tm), "activate", G_CALLBACK(cdsdwindow_close), NULL);
		g_signal_connect(G_OBJECT(cdsdwindow.trayicon), "activate", G_CALLBACK(cdsdwindow_trayaction), NULL);
		
		gtk_menu_shell_append(GTK_MENU_SHELL(cdsdwindow.traymenu), cdsdwindow.showwin_tm);
		gtk_menu_shell_append(GTK_MENU_SHELL(cdsdwindow.traymenu), cdsdwindow.sep1_tm);
		gtk_menu_shell_append(GTK_MENU_SHELL(cdsdwindow.traymenu), cdsdwindow.devices_tm);
		gtk_menu_shell_append(GTK_MENU_SHELL(cdsdwindow.traymenu), cdsdwindow.sep2_tm);
		gtk_menu_shell_append(GTK_MENU_SHELL(cdsdwindow.traymenu), cdsdwindow.quit_tm);
		
		cdsdwindow.devicesmenu = gtk_menu_new();
		
		gtk_menu_item_set_submenu(GTK_MENU_ITEM(cdsdwindow.devices_tm), cdsdwindow.devicesmenu);
		
		cdsdwindow.speedsmenu = gtk_menu_new();
		cdsdwindow.cdspeeds_tm = gtk_menu_item_new_with_label(_("CD speed"));
		cdsdwindow.dvdspeeds_tm = gtk_menu_item_new_with_label(_("DVD speed"));
		cdsdwindow.bdspeeds_tm = gtk_menu_item_new_with_label(_("BD speed"));
		
		gtk_menu_shell_append(GTK_MENU_SHELL(cdsdwindow.speedsmenu), cdsdwindow.cdspeeds_tm);
		gtk_menu_shell_append(GTK_MENU_SHELL(cdsdwindow.speedsmenu), cdsdwindow.dvdspeeds_tm);
		gtk_menu_shell_append(GTK_MENU_SHELL(cdsdwindow.speedsmenu), cdsdwindow.bdspeeds_tm);
		
		cdsdwindow.cdspeedsmenu = gtk_menu_new();
		i = 0;
		while (cdspeeds[i] != -1) {
			if (cdspeeds[i] == 0) {
				cdsdwindow.cdspeed_tm[i] = gtk_radio_menu_item_new_with_label(group, _("Optimal"));
			} else {
				snprintf((gchar *)&textbuf, sizeof(textbuf), "%u x", cdspeeds[i]);
				cdsdwindow.cdspeed_tm[i] = gtk_radio_menu_item_new_with_label(group, (gchar *)&textbuf);
			}
			
			g_signal_connect(G_OBJECT(cdsdwindow.cdspeed_tm[i]), "activate", G_CALLBACK(cdsdwindow_device_set_speed_tray), GINT_TO_POINTER(CDSD_SPEED_TRAY_PACK(TRAY_CDSPEED, i)));
			
			gtk_menu_shell_append(GTK_MENU_SHELL(cdsdwindow.cdspeedsmenu), cdsdwindow.cdspeed_tm[i]);
			group = gtk_radio_menu_item_get_group(GTK_RADIO_MENU_ITEM(cdsdwindow.cdspeed_tm[i]));
			i++;
		}
		
		gtk_menu_item_set_submenu(GTK_MENU_ITEM(cdsdwindow.cdspeeds_tm), cdsdwindow.cdspeedsmenu);
		
		group = NULL;
		
		cdsdwindow.dvdspeedsmenu = gtk_menu_new();
		i = 0;
		while (dvdspeeds[i] != -1) {
			if (dvdspeeds[i] == 0) {
				cdsdwindow.dvdspeed_tm[i] = gtk_radio_menu_item_new_with_label(group, _("Optimal"));
			} else {
				snprintf((gchar *)&textbuf, sizeof(textbuf), "%u x", dvdspeeds[i]);
				cdsdwindow.dvdspeed_tm[i] = gtk_radio_menu_item_new_with_label(group, (gchar *)&textbuf);
			}
			
			g_signal_connect(G_OBJECT(cdsdwindow.dvdspeed_tm[i]), "activate", G_CALLBACK(cdsdwindow_device_set_speed_tray), GINT_TO_POINTER(CDSD_SPEED_TRAY_PACK(TRAY_DVDSPEED, i)));
			
			gtk_menu_shell_append(GTK_MENU_SHELL(cdsdwindow.dvdspeedsmenu), cdsdwindow.dvdspeed_tm[i]);
			group = gtk_radio_menu_item_get_group(GTK_RADIO_MENU_ITEM(cdsdwindow.dvdspeed_tm[i]));
			i++;
		}
		
		gtk_menu_item_set_submenu(GTK_MENU_ITEM(cdsdwindow.dvdspeeds_tm), cdsdwindow.dvdspeedsmenu);
		
		group = NULL;
		
		cdsdwindow.bdspeedsmenu = gtk_menu_new();
		i = 0;
		while (bdspeeds[i] != -1) {
			if (bdspeeds[i] == 0) {
				cdsdwindow.bdspeed_tm[i] = gtk_radio_menu_item_new_with_label(group, _("Optimal"));
			} else {
				snprintf((gchar *)&textbuf, sizeof(textbuf), "%u x", bdspeeds[i]);
				cdsdwindow.bdspeed_tm[i] = gtk_radio_menu_item_new_with_label(group, (gchar *)&textbuf);
			}
			
			g_signal_connect(G_OBJECT(cdsdwindow.bdspeed_tm[i]), "activate", G_CALLBACK(cdsdwindow_device_set_speed_tray), GINT_TO_POINTER(CDSD_SPEED_TRAY_PACK(TRAY_BDSPEED, i)));
			
			gtk_menu_shell_append(GTK_MENU_SHELL(cdsdwindow.bdspeedsmenu), cdsdwindow.bdspeed_tm[i]);
			group = gtk_radio_menu_item_get_group(GTK_RADIO_MENU_ITEM(cdsdwindow.bdspeed_tm[i]));
			i++;
		}
		
		gtk_menu_item_set_submenu(GTK_MENU_ITEM(cdsdwindow.bdspeeds_tm), cdsdwindow.bdspeedsmenu);
				
		gtk_widget_show_all(cdsdwindow.traymenu);
		
		g_signal_connect(G_OBJECT(cdsdwindow.trayicon), "popup-menu", G_CALLBACK(cdsdwindow_traypopup), G_OBJECT(cdsdwindow.traymenu));
		
	} else {
		gtk_window_present(GTK_WINDOW(cdsdwindow.win));
	}
	
	return cdsdwindow.win;
}

static int cdsdctl_load_library(void)
{
	void *cdsdlib;
	char *liberror;
	int i;
	
	typedef struct {
		void **function;
		const char *alias;
	} library;
	
	library cdsdcore[] = {
		{(void **)(&libcdsd_command_list_dl), "libcdsd_command_list"},
		{(void **)(&libcdsd_command_modeset_dl), "libcdsd_command_modeset"},
		{(void **)(&libcdsd_command_info_dl), "libcdsd_command_info"},
		{(void **)(&libcdsd_command_capabilities_dl), "libcdsd_command_capabilities"},
		{(void **)(&libcdsd_connection_available_dl), "libcdsd_connection_available"},
		/*{(void **)(&libcdsd_get_error_dl), "libcdsd_get_error"},*/
		{NULL, NULL}
	};
	
	
	if ((cdsdlib = dlopen(RESOURCE_CDSD_LIBRARY, RTLD_LAZY)) == NULL) {
		fprintf(stderr, "Failed to open shared library: %s\n", dlerror());
		return -1;
	}
	
	dlerror();
	
	i = 0;
	while (cdsdcore[i].alias != NULL) {
		*(cdsdcore[i].function) = dlsym(cdsdlib, cdsdcore[i].alias); 
		if ((liberror = dlerror()) != NULL) {
			fprintf(stderr, "Failed to assign library function \"%s\": %s\n", cdsdcore[i].alias, liberror);
			return -1;
		}
		i++;
	}
	
	return 0;
}

static void cdsdwindow_init_lib(void)
{
	int error;
	
	if (cdsdctl_load_library() == -1) {
		cdsdwindow_message(_("Unable to load shared library"), _("CDSD error"), GTK_MESSAGE_ERROR, NULL);
		exit(EXIT_FAILURE);
	}
	
	if (libcdsd_connection_available_dl(&error) == -1) {
		cdsdwindow_message(_("CDSD service not running"), _("CDSD error"), GTK_MESSAGE_ERROR, NULL);
		exit(EXIT_FAILURE);
	}
}

static void cdsdwindow_request_device_list(void)
{
	int error;
	
	if (libcdsd_command_list_dl(&error, cdsdwindow_add_event_cb) < 1) {
		switch (error) {
			case CDSD_EINTERNAL:
				cdsdwindow_message(_("Internal error"), _("CDSD error"), GTK_MESSAGE_ERROR, NULL);
				break;
			case CDSD_EACCESS:
				cdsdwindow_message(_("No access permission\nAre you memeber of 'cdrom' group?\nTry running 'usermod -a -G cdrom <username>' as root."), _("CDSD error"), GTK_MESSAGE_ERROR, NULL);
				break;
			case CDSD_ENODEVICES:
				cdsdwindow_message(_("No devices available"), _("CDSD error"), GTK_MESSAGE_ERROR, NULL);
				break;
			case CDSD_ECOMMANDFAIL:
				cdsdwindow_message(_("Command execution fail"), _("CDSD error"), GTK_MESSAGE_ERROR, NULL);
				break;
			case CDSD_ENOSERVICE:
				cdsdwindow_message(_("Service not running"), _("CDSD error"), GTK_MESSAGE_ERROR, NULL);
				break;
			case CDSD_EREQUESTFAIL:
				cdsdwindow_message(_("Wrong request"), _("CDSD error"), GTK_MESSAGE_ERROR, NULL);
				break;
			default:
				cdsdwindow_message(_("Unknown error"), _("CDSD error"), GTK_MESSAGE_ERROR, NULL);
				break;
		}
		exit(EXIT_FAILURE);
	}
}

#if GTK_CHECK_VERSION(3, 0, 0)
static void cdsdwindow_activate(GtkApplication *app)
{
	GList *list;
	GtkWidget *window;
	
	list = gtk_application_get_windows(app);
	
	if (list) {
		gtk_window_present(GTK_WINDOW(g_list_nth(list, 0)));
	} else {
		cdsdwindow_init_lib();
		window = cdsdwindow_create();
		cdsdwindow_request_device_list();
		gtk_window_set_application(GTK_WINDOW(window), app);
		gtk_widget_show_all(window);
	}
}
#endif

int main (int argc, char *argv[])
{
	gint status;
	#if GTK_CHECK_VERSION(3, 0, 0)
		GtkApplication *app;
	#endif
	
	#ifndef LC_ALL
		#define LC_ALL 0
	#endif
	
	setlocale(LC_ALL, "");
	bindtextdomain(CDSDGUI_GETTEXT_PACKAGE, RESOURCE_DIR_LOCALE);
	bind_textdomain_codeset(CDSDGUI_GETTEXT_PACKAGE, "UTF-8");
	textdomain(CDSDGUI_GETTEXT_PACKAGE);
	
	g_set_application_name("CDSD drive speed control");
	
	
	#if GTK_CHECK_VERSION(3, 0, 0)
		app = gtk_application_new("org.gtk.CDSD", 0);
		g_signal_connect(app, "activate", G_CALLBACK(cdsdwindow_activate), NULL);
	#else
		gtk_init(&argc, &argv);
		cdsdwindow_init_lib();
		cdsdwindow_create();
		cdsdwindow_request_device_list();
	#endif
	
	status = 0;
	
	#if GTK_CHECK_VERSION(3, 0, 0)
		status = g_application_run(G_APPLICATION(app), argc, argv);
		g_object_unref(app);
	#else
		gtk_main();
	#endif
	
	return status;
}
