/*
 *      cdsd.h
 *      
 *      Copyright 2011 Alex <alex@linuxonly.ru>
 *      
 *      This program is free software: you can redistribute it and/or modify
 *      it under the terms of the GNU General Public License as published by
 *      the Free Software Foundation; either version 3 of the License, or
 *      (at your option) any later version.
 *      
 *      This program is distributed in the hope that it will be useful,
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of
 *      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *      GNU General Public License for more details.
 *      
 *      You should have received a copy of the GNU General Public License
 *      along with this program. If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef __CDSD_H__
#define __CDSD_H__

#include <limits.h>
#include <grp.h>

#define MTAB_PATH_ETC     "/etc/mtab"
#define MTAB_PATH_TMP     "/var/run/mtab.tmp"
#define MTAB_PATH_ESN     "/etc/mtab.tmp"

struct optical_device {
	//Device parameters
	unsigned int id;
	int fd;
	char *devpath;
	unsigned int major;
	unsigned int minor;
	char vendorid[9];
	char productid[17];
	char productver[5];
	char productsn[21];
	//Media type string
	char mediastr[32];
	unsigned int media_profile;
	unsigned int capabilities_profile;
	unsigned int media_sessions_number;
	unsigned long int media_end_lba;
	unsigned long int media_block_length;
	//Device speed values
	unsigned short buffer_size;
	unsigned short max_read_speed;
	unsigned short cur_read_speed;
	unsigned short max_write_speed;
	unsigned short cur_write_speed;
	//User-defined speed values
	unsigned int user_cd_speed;
	unsigned int user_dvd_speed;
	unsigned int user_bd_speed;
	unsigned int user_polling;
	unsigned int user_state_modified : 1;
	//Event flags
	unsigned event_eject_request     : 1;
	unsigned event_new_media         : 1;
	unsigned event_media_removal     : 1;
	unsigned event_media_changed     : 1;
	unsigned event_bgformat_completed: 1;
	unsigned event_bgformat_restarted: 1;
	unsigned event_tray_open         : 1;
	unsigned event_media_present     : 1;
	//Media type flags
	unsigned media_disc_status       : 2;
	unsigned media_session_state     : 2;
	unsigned media_bg_format_status  : 2;
	unsigned media_erasable          : 1;
	unsigned media_exists            : 1;
	unsigned media_type_cd           : 1;
	unsigned media_type_dvd          : 1;
	unsigned media_type_bd           : 1;
	unsigned media_mo                : 1;
	unsigned media_cd_rom            : 1;
	unsigned media_cd_r              : 1;
	unsigned media_cd_rw             : 1;
	unsigned media_dvd_rom           : 1;
	unsigned media_dvd_r             : 1;
	unsigned media_dvd_ram           : 1;
	unsigned media_dvd_rw            : 1;
	unsigned media_dvd_rw_ro         : 1;
	unsigned media_dvd_rw_seq        : 1;
	unsigned media_dvd_r_dl_seq      : 1;
	unsigned media_dvd_r_dl_jmp      : 1;
	unsigned media_dvd_rw_dl         : 1;
	unsigned media_dvd_plus_r        : 1;
	unsigned media_dvd_plus_rw       : 1;
	unsigned media_dvd_plus_rw_dl    : 1;
	unsigned media_dvd_plus_r_dl     : 1;
	unsigned media_bd                : 1;
	unsigned media_bd_r              : 1;
	unsigned media_bd_re             : 1;
	unsigned media_hddvd             : 1;
	unsigned media_hddvd_r           : 1;
	unsigned media_hddvd_rw          : 1;
	unsigned media_hddvd_r_dl        : 1;
	unsigned media_hddvd_rw_dl       : 1;
	//Device capabilities flags
	//from MODE SENSE SCSI command
	unsigned cap_dvd_ram_read        : 1;
	unsigned cap_dvd_ram_write       : 1;
	unsigned cap_dvd_r_read          : 1;
	unsigned cap_dvd_r_write         : 1;
	unsigned cap_dvd_rom_read        : 1;
	unsigned cap_simulate            : 1;
	unsigned cap_cd_rw_read          : 1;
	unsigned cap_cd_rw_write         : 1;
	unsigned cap_cd_r_read           : 1;
	unsigned cap_cd_r_write          : 1;
	//from GET CONFIGURATION SCSI command
	unsigned cap_mo_read             : 1;
	unsigned cap_cd_rom_read         : 1;
	unsigned cap_dvd_rw_read         : 1;
	unsigned cap_dvd_rw_ro_read      : 1;
	unsigned cap_dvd_rw_seq_read     : 1;
	unsigned cap_dvd_r_dl_read       : 1;
	unsigned cap_dvd_rw_dl_read      : 1;
	unsigned cap_dvd_plus_r_read     : 1;
	unsigned cap_dvd_plus_rw_read    : 1;
	unsigned cap_dvd_plus_rw_dl_read : 1;
	unsigned cap_dvd_plus_r_dl_read  : 1;
	unsigned cap_bd_read             : 1;
	unsigned cap_bd_r_read           : 1;
	unsigned cap_bd_re_read          : 1;
	unsigned cap_hddvd_read          : 1;
	unsigned cap_hddvd_r_read        : 1;
	unsigned cap_hddvd_rw_read       : 1;
	unsigned cap_hddvd_r_dl_read     : 1;
	unsigned cap_hddvd_rw_dl_read    : 1;
};

struct cdsd_global_settings {
	gid_t validgroup;
	char validgroupname[256];
	unsigned int online_speed_change;
	unsigned int default_cd_speed;
	unsigned int default_dvd_speed;
	unsigned int default_bd_speed;
	unsigned int default_polling;
};

int cdsd_event_cb(unsigned int major, unsigned int minor);
int cdsd_scan_devices(struct cdsd_global_settings *settings);

#endif /* __CDSD_H__ */
