/*
 *      datalist.c
 *      
 *      Copyright 2009 - 2011 Alex <alex@linuxonly.ru>
 *      
 *      This program is free software: you can redistribute it and/or modify
 *      it under the terms of the GNU General Public License as published by
 *      the Free Software Foundation; either version 3 of the License, or
 *      (at your option) any later version.
 *      
 *      This program is distributed in the hope that it will be useful,
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of
 *      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *      GNU General Public License for more details.
 *      
 *      You should have received a copy of the GNU General Public License
 *      along with this program. If not, see <http://www.gnu.org/licenses/>.
 */


#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "datalist.h"

int datalist_get_empty(nodeptr pointer)
{
	return (pointer == NULL);
}

unsigned int datalist_get_length(nodeptr pointer)
{
	unsigned int length = 0;
	
	while (pointer != NULL) {
		pointer = pointer->nextptr;
		length++;
	}
	
	return length;
}

int datalist_add_node(nodeptr *pointer, char *data)
{
	nodeptr previous, current;
	nodeptr newnode = calloc(1, sizeof(nodeptr));
	unsigned int index = 0;
	
	if (newnode == NULL) return -1;
	
	newnode->data = data;
	newnode->nextptr = NULL;
	
	previous = NULL;
	current = *pointer;
	
	while (current != NULL) {
		previous = current;
		current = current->nextptr;
		index++;
	}
	
	if (previous == NULL) {
		newnode->nextptr = *pointer;
		*pointer = newnode;
		
		return 0;
	} else {
		previous->nextptr = newnode;
		newnode->nextptr = NULL;
		
		return index;
	}
}

int datalist_insert_node(nodeptr *pointer, char *data, unsigned int dindex)
{
	nodeptr previous, current;
	nodeptr newnode = calloc(1, sizeof(nodeptr));
	unsigned int index = 0;
	
	if (newnode == NULL) return -1;
	
	newnode->data = data;
	newnode->nextptr = NULL;
		
	previous = NULL;
	current = *pointer;
	
	while ((current != NULL) && (index != dindex)) {
		previous = current;
		current = current->nextptr;
		index++;
	}
	
	if (previous == NULL) {
		newnode->nextptr = *pointer;
		*pointer = newnode;
		return 0;
	} else {
		previous->nextptr = newnode;
		newnode->nextptr = current;
		return index;
	}
}

int datalist_remove_node(nodeptr *pointer, unsigned int dindex)
{
	nodeptr previous, current;
	unsigned int index = 0;
	
	if (*pointer == NULL) return -1;
	
	if (dindex == index) {
		previous = *pointer;
		*pointer = (*pointer)->nextptr;
		free(previous->data);
		free(previous);

		return index;
	}
	
	previous = *pointer;
	current = (*pointer)->nextptr;
	index++;
	
	while (current->nextptr != NULL) {
		if (dindex == index) {
			previous->nextptr = current->nextptr;
			free(current->data);
			free(current);

			return index;
		}
		previous = current;
		current = current->nextptr;
		index++;
	}
	
	previous->nextptr = NULL;
	free(current->data);
	free(current);

	return index;
}

void datalist_foreach(nodeptr list, void *userdata, int(*callback(void **data, int dindex, void *userdata)))
{
	if ((callback == NULL) || (list == NULL)) return;
	int index = 0;
	
	while (list != NULL) {
		if (callback(&(list->data), index, userdata)) return;
		list = list->nextptr;
		index++;
	}
}

void *datalist_get_nth(nodeptr list, unsigned int n)
{
	if (list == NULL) return NULL;
	int index = 0;
	
	while (list != NULL) {
		if (index == n) return list->data;
		list = list->nextptr;
		index++;
	}
	
	return NULL;
}

void datalist_free(nodeptr *list)
{
	nodeptr nextptr, ptr;
	
	if (*list == NULL) return;
	
	ptr = *list;
	nextptr = (*list)->nextptr;
	free(ptr->data);
	free(ptr);
	while (nextptr != NULL) {
		ptr = nextptr;
		nextptr = nextptr->nextptr;
		free(ptr->data);
		free(ptr);
	}
	*list = NULL;
}

void datalist_free_recursive(nodeptr *list, void *userdata, int(*callback(void **data, int dindex, void *userdata)))
{
	nodeptr nextptr, ptr;
	int ind = 0;
	
	if ((*list == NULL) || (callback == NULL)) return;
	
	ptr = *list;
	nextptr = (*list)->nextptr;
	callback(&(ptr->data), ind, userdata);
	free(ptr->data);
	free(ptr);
	ind++;
	while (nextptr != NULL) {
		ptr = nextptr;
		nextptr = nextptr->nextptr;
		callback(&(ptr->data), ind, userdata);
		free(ptr->data);
		free(ptr);
		ind++;
	}
	*list = NULL;
}
