/*
 *      libcdsd.h
 *      
 *      Copyright 2011 Alex <alex@linuxonly.ru>
 *      
 *      This program is free software: you can redistribute it and/or modify
 *      it under the terms of the GNU General Public License as published by
 *      the Free Software Foundation; either version 3 of the License, or
 *      (at your option) any later version.
 *      
 *      This program is distributed in the hope that it will be useful,
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of
 *      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *      GNU General Public License for more details.
 *      
 *      You should have received a copy of the GNU General Public License
 *      along with this program. If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef __LIBCDSD_H__
#define __LIBCDSD_H__

#include <stdio.h>

#define CDSD_ANSWER_TIMEOUT  15000

#define CDSD_ESUCCESS       1
#define CDSD_EINTERNAL      2
#define CDSD_EUNKNOWN       3
#define CDSD_EACCESS        4
#define CDSD_ENODEVICES     5
#define CDSD_ECOMMANDFAIL   6
#define CDSD_ENOSERVICE     7
#define CDSD_EREQUESTFAIL   8

struct cdsd_settings_device {
	char vendor[9];
	char product[17];
	char fname[17];
	unsigned int id;
	unsigned int media;
	unsigned int capabilities;
	unsigned int cdspeed;
	unsigned int dvdspeed;
	unsigned int bdspeed;
	unsigned int polling;
};

struct cdsd_settings_info {
	char vendorid[9];
	char productid[17];
	char productver[5];
	char productsn[21];
	char media_exists;
	char media_erasable;
	char media_type_cd;
	char media_type_dvd;
	char media_type_bd;
	unsigned short buffer_size;
	unsigned short max_read_speed;
	unsigned short cur_read_speed;
	unsigned short max_write_speed;
	unsigned short cur_write_speed;
	
};

struct cdsd_capabilities_info {
	unsigned char cap_cdrom;
	unsigned char cap_cd_r;
	unsigned char cap_cd_rw;
	unsigned char cap_dvd_rom;
	unsigned char cap_dvd_ram;
	unsigned char cap_dvd_r;
	unsigned char cap_dvd_r_dl;
	unsigned char cap_dvd_rw;
	unsigned char cap_dvd_rw_dl;
	unsigned char cap_dvd_plus_r;
	unsigned char cap_dvd_plus_r_dl;
	unsigned char cap_dvd_plus_rw;
	unsigned char cap_dvd_plus_rw_dl;
	unsigned char cap_bd_rom;
	unsigned char cap_bd_r;
	unsigned char cap_bd_re;
	unsigned char cap_hddvd_rom;
	unsigned char cap_hddvd_r;
	unsigned char cap_hddvd_r_dl;
	unsigned char cap_hddvd_rw;
	unsigned char cap_hddvd_rw_dl;
};

typedef int (*libcdsd_list_cb)(struct cdsd_settings_device *device);

void _init(void);
void _fini(void);

int libcdsd_command_list(int *error, libcdsd_list_cb list_cb);
int libcdsd_command_modeset(int *error, struct cdsd_settings_device *device);
int libcdsd_command_info(int *error, int deviceid, struct cdsd_settings_info *info);
int libcdsd_connection_available(int *error);
//int libcdsd_get_error(void);

#endif /* __LIBCDSD_H__ */
