/*
 *      scsi.c
 *      
 *      Copyright 2011 Alex <alex@linuxonly.ru>
 *      
 *      This program is free software: you can redistribute it and/or modify
 *      it under the terms of the GNU General Public License as published by
 *      the Free Software Foundation; either version 3 of the License, or
 *      (at your option) any later version.
 *      
 *      This program is distributed in the hope that it will be useful,
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of
 *      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *      GNU General Public License for more details.
 *      
 *      You should have received a copy of the GNU General Public License
 *      along with this program. If not, see <http://www.gnu.org/licenses/>.
 */

#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <fcntl.h>
#include <string.h>
#include <errno.h>
#include <sys/ioctl.h>

#include <scsi/scsi_ioctl.h>
#include <scsi/sg.h>

#include "cdsd.h"
#include "scsi.h"

static void scsi_strip_whitespaces(char *string, size_t size);


int scsi_profile_set(int profile, int action, struct optical_device *device)
{
	switch (profile) {
		case 0x03:
		case 0x04:
		case 0x05:
			if (action == MMC_PROFILE_MEDIA) {
				snprintf((char *)&device->mediastr, sizeof(device->mediastr), "MO erasable");
				device->media_exists = 1;
				device->media_mo = 1;
			} else if (action == MMC_PROFILE_CAP) {
				device->cap_mo_read = 1;
			}
			break;
		case 0x08:
			if (action == MMC_PROFILE_MEDIA) {
				snprintf((char *)&device->mediastr, sizeof(device->mediastr), "CD-ROM");
				device->media_exists = 1;
				device->media_cd_rom = 1;
				device->media_type_cd =	1;
			} else if (action == MMC_PROFILE_CAP) {
				device->cap_cd_rom_read = 1;
			}
			break;
		case 0x09:
			if (action == MMC_PROFILE_MEDIA) {
				snprintf((char *)&device->mediastr, sizeof(device->mediastr), "CD-R");
				device->media_exists = 1;
				device->media_cd_r = 1;
				device->media_type_cd =	1;
			} else if (action == MMC_PROFILE_CAP) {
				device->cap_cd_r_read = 1;
			}
			break;
		case 0x0a:
			if (action == MMC_PROFILE_MEDIA) {
				snprintf((char *)&device->mediastr, sizeof(device->mediastr), "CD-RW");
				device->media_exists = 1;
				device->media_cd_rw = 1;
				device->media_type_cd =	1;
			} else if (action == MMC_PROFILE_CAP) {
				device->cap_cd_rw_read = 1;
			}
			break;
		case 0x10:
			if (action == MMC_PROFILE_MEDIA) {
				snprintf((char *)&device->mediastr, sizeof(device->mediastr), "DVD-ROM");
				device->media_exists = 1;
				device->media_dvd_rom = 1;
				device->media_type_dvd = 1;
			} else if (action == MMC_PROFILE_CAP) {
				device->cap_dvd_rom_read = 1;
			}
			break;
		case 0x11:
			if (action == MMC_PROFILE_MEDIA) {
				snprintf((char *)&device->mediastr, sizeof(device->mediastr), "DVD-R sequential recording");
				device->media_exists = 1;
				device->media_dvd_r = 1;
				device->media_type_dvd = 1;
			} else if (action == MMC_PROFILE_CAP) {
				device->cap_dvd_r_read = 1;
			}
			break;
		case 0x12:
			if (action == MMC_PROFILE_MEDIA) {
				snprintf((char *)&device->mediastr, sizeof(device->mediastr), "DVD-RAM");
				device->media_exists = 1;
				device->media_dvd_ram = 1;
				device->media_type_dvd = 1;
			} else if (action == MMC_PROFILE_CAP) {
				device->cap_dvd_ram_read = 1;
			}
			break;
		case 0x13:
			if (action == MMC_PROFILE_MEDIA) {
				snprintf((char *)&device->mediastr, sizeof(device->mediastr), "DVD-RW restricted overwrite");
				device->media_exists = 1;
				device->media_dvd_rw = 1;
				device->media_dvd_rw_ro = 1;
				device->media_type_dvd = 1;
			} else if (action == MMC_PROFILE_CAP) {
				device->cap_dvd_rw_read = 1;
			}
			break;
		case 0x14:
			if (action == MMC_PROFILE_MEDIA) {
				snprintf((char *)&device->mediastr, sizeof(device->mediastr), "DVD-RW sequential recording");
				device->media_exists = 1;
				device->media_dvd_rw = 1;
				device->media_dvd_rw_seq = 1;
				device->media_type_dvd = 1;
			} else if (action == MMC_PROFILE_CAP) {
				device->cap_dvd_rw_read = 1;
				device->cap_dvd_rw_seq_read = 1;
			}
			break;
		case 0x15:
			if (action == MMC_PROFILE_MEDIA) {
				snprintf((char *)&device->mediastr, sizeof(device->mediastr), "Dual Layer DVD-R using Sequential recording");
				device->media_exists = 1;
				device->media_dvd_r_dl_seq = 1;
				device->media_type_dvd = 1;
			} else if (action == MMC_PROFILE_CAP) {
				device->cap_dvd_r_dl_read = 1;
			}
			break;
		case 0x16:
			if (action == MMC_PROFILE_MEDIA) {
				snprintf((char *)&device->mediastr, sizeof(device->mediastr), "Dual Layer DVD-R using Layer Jump recording");
				device->media_exists = 1;
				device->media_dvd_r_dl_jmp = 1;
				device->media_type_dvd = 1;
			} else if (action == MMC_PROFILE_CAP) {
				device->cap_dvd_r_dl_read = 1;
			}
			break;
		case 0x17:
			if (action == MMC_PROFILE_MEDIA) {
				snprintf((char *)&device->mediastr, sizeof(device->mediastr), "Dual Layer DVD-RW recording");
				device->media_exists = 1;
				device->media_dvd_rw_dl = 1;
				device->media_type_dvd = 1;
			} else if (action == MMC_PROFILE_CAP) {
				device->cap_dvd_rw_dl_read = 1;
			}
			break;	
		case 0x1B:
			if (action == MMC_PROFILE_MEDIA) {
				snprintf((char *)&device->mediastr, sizeof(device->mediastr), "DVD+R");
				device->media_exists = 1;
				device->media_dvd_plus_r = 1;
				device->media_type_dvd = 1;
			} else if (action == MMC_PROFILE_CAP) {
				device->cap_dvd_plus_r_read = 1;
			}
			break;
		case 0x1A:
			if (action == MMC_PROFILE_MEDIA) {
				snprintf((char *)&device->mediastr, sizeof(device->mediastr), "DVD+RW");
				device->media_exists = 1;
				device->media_dvd_plus_rw = 1;
				device->media_type_dvd = 1;
			} else if (action == MMC_PROFILE_CAP) {
				device->cap_dvd_plus_rw_read = 1;
			}
			break;
		case 0x2A:
			if (action == MMC_PROFILE_MEDIA) {
				snprintf((char *)&device->mediastr, sizeof(device->mediastr), "DVD+RW/DL");
				device->media_exists = 1;
				device->media_dvd_plus_rw_dl = 1;
				device->media_type_dvd = 1;
			} else if (action == MMC_PROFILE_CAP) {
				device->cap_dvd_plus_rw_dl_read = 1;
			}
			break;
		case 0x2B:
			if (action == MMC_PROFILE_MEDIA) {
				snprintf((char *)&device->mediastr, sizeof(device->mediastr), "DVD+R/DL");
				device->media_exists = 1;
				device->media_dvd_plus_r_dl = 1;
				device->media_type_dvd = 1;
			} else if (action == MMC_PROFILE_CAP) {
				device->cap_dvd_plus_r_dl_read = 1;
			}
			break;
		case 0x40:
			if (action == MMC_PROFILE_MEDIA) {
				snprintf((char *)&device->mediastr, sizeof(device->mediastr), "BD-ROM");
				device->media_exists = 1;
				device->media_bd = 1;
				device->media_type_bd = 1;
			} else if (action == MMC_PROFILE_CAP) {
				device->cap_bd_read = 1;
			}
			break;
		case 0x41:
		case 0x42:
			if (action == MMC_PROFILE_MEDIA) {
				snprintf((char *)&device->mediastr, sizeof(device->mediastr), "BD-R random recording");
				device->media_exists = 1;
				device->media_bd_r = 1;
				device->media_type_bd = 1;
			} else if (action == MMC_PROFILE_CAP) {
				device->cap_bd_r_read = 1;
			}
			break;
		case 0x43:
			if (action == MMC_PROFILE_MEDIA) {
				snprintf((char *)&device->mediastr, sizeof(device->mediastr), "BD-RE");
				device->media_exists = 1;
				device->media_bd_re = 1;
				device->media_type_bd = 1;
			} else if (action == MMC_PROFILE_CAP) {
				device->cap_bd_re_read = 1;
			}
			break;
		case 0x50:
			if (action == MMC_PROFILE_MEDIA) {
				snprintf((char *)&device->mediastr, sizeof(device->mediastr), "HD-DVD-ROM");
				device->media_exists = 1;
				device->media_hddvd = 1;
				device->media_type_bd = 1;
			} else if (action == MMC_PROFILE_CAP) {
				device->cap_hddvd_read = 1;
			}
			break;
		case 0x51:
			if (action == MMC_PROFILE_MEDIA) {
				snprintf((char *)&device->mediastr, sizeof(device->mediastr), "HD-DVD-R");
				device->media_exists = 1;
				device->media_hddvd_r = 1;
				device->media_type_bd = 1;
			} else if (action == MMC_PROFILE_CAP) {
				device->cap_hddvd_r_read = 1;
			}
			break;
		case 0x52:
			if (action == MMC_PROFILE_MEDIA) {
				snprintf((char *)&device->mediastr, sizeof(device->mediastr), "HD-DVD-RAM");
				device->media_exists = 1;
				device->media_hddvd_rw = 1;
				device->media_type_bd = 1;
			} else if (action == MMC_PROFILE_CAP) {
				device->cap_hddvd_rw_read = 1;
			}
			break;
		case 0x58:
			if (action == MMC_PROFILE_MEDIA) {
				snprintf((char *)&device->mediastr, sizeof(device->mediastr), "Dual Layer Write-once HD DVD");
				device->media_exists = 1;
				device->media_hddvd_r_dl = 1;
				device->media_type_bd = 1;
			} else if (action == MMC_PROFILE_CAP) {
				device->cap_hddvd_r_dl_read = 1;
			}
			break;
		case 0x5A:
			if (action == MMC_PROFILE_MEDIA) {
				snprintf((char *)&device->mediastr, sizeof(device->mediastr), "Dual Layer Rewritable HD DVD");
				device->media_exists = 1;
				device->media_hddvd_rw_dl = 1;
				device->media_type_bd = 1;
			} else if (action == MMC_PROFILE_CAP) {
				device->cap_hddvd_rw_dl_read = 1;
			}
			break;
		default:
			printf("[-] Media profile 0x%02x unknown\n", profile);
			return 0;
					
	}
	
	return 1;		
}

static void scsi_strip_whitespaces(char *string, size_t size)
{
	int p;
	
	for (p = size - 2; p >= 0; p--) {
		if (string[p] == 0x20) string[p] = 0; else break;
	}
}

int scsi_test_device_writeable(int fd)
{
	int fdflags;
	
	fdflags = fcntl(fd, F_GETFD, 0);
	
	if ((fdflags == -1) || (fdflags & O_RDONLY)) return 0;
	
	return 1;
}

int scsi_test_device_media_blank(struct optical_device *device)
{
	if (device == NULL) return 1;
	
	if (device->media_disc_status == 0) return 1;
	
	return 0;
}

int scsi_test_unit_ready(int fd, unsigned int timeout)
{
	struct sg_io_hdr header;
	unsigned char cmdbuf[6];
	int result = 0;
	
	memset(&header, 0, sizeof(struct sg_io_hdr));
	header.interface_id = 'S';
	header.flags = SG_FLAG_LUN_INHIBIT | SG_FLAG_DIRECT_IO;
	header.dxfer_direction = SG_DXFER_NONE;
	header.cmdp = cmdbuf;
	header.cmd_len = sizeof(cmdbuf);
	header.dxferp = NULL;
	header.dxfer_len = 0;
	header.sbp = NULL;
	header.mx_sb_len = 0;
	header.timeout = timeout;
	
	memset(cmdbuf, 0, sizeof(cmdbuf));
	
	if ((result = ioctl(fd, SG_IO, &header)) < 0) {
		printf("[-] Failed to send SCSI TEST UNIT command\n");
		return result;
	}
	
	if (header.status == 0) result = 1;
	
	return result;
}

int scsi_inquiry(int fd, int page_code, int evpd, struct optical_device *device)
{
	struct sg_io_hdr header;
	unsigned char recvdata[255];
	unsigned char cmdbuf[6];
	int result = 0;
	
	memset(&header, 0, sizeof(struct sg_io_hdr));
	header.interface_id = 'S';
	header.flags = SG_FLAG_LUN_INHIBIT | SG_FLAG_DIRECT_IO;
	header.dxfer_direction = SG_DXFER_FROM_DEV;
	header.cmdp = cmdbuf;
	header.cmd_len = sizeof(cmdbuf);
	header.dxferp = recvdata;
	header.dxfer_len = sizeof(recvdata);
	header.sbp = NULL;
	header.mx_sb_len = 0;
	
	memset(&cmdbuf, 0, sizeof(cmdbuf));
	cmdbuf[0] = 0x12;
	cmdbuf[1] = evpd & 1;
	cmdbuf[2] = page_code & 0xff;
	cmdbuf[3] = 0;
	cmdbuf[4] = 0xff;
	cmdbuf[5] = 0;
	
	if ((result = ioctl(fd, SG_IO, &header)) < 0) {
		printf("[-] Failed to send SCSI INQUIRY command\n");
		return result;
	}
	
	if (header.status == 0) {
		if (((recvdata[0] & 0x1f) == 0x05) || ((recvdata[0] & 0x1f) == 0x07)) {
			memcpy(&device->vendorid, &recvdata[8], 8);
			scsi_strip_whitespaces((char *)&device->vendorid, sizeof(device->vendorid));
			memcpy(&device->productid, &recvdata[16], 16);
			scsi_strip_whitespaces((char *)&device->productid, sizeof(device->productid));
			memcpy(&device->productver, &recvdata[32], 4);
			scsi_strip_whitespaces((char *)&device->productver, sizeof(device->productver));
			memcpy(&device->productsn, &recvdata[36], 20);
			scsi_strip_whitespaces((char *)&device->productsn, sizeof(device->productsn));
			result = 1;
		}
	}
	
	return result;
}

int scsi_get_configuration(int fd, struct optical_device *device)
{
	struct sg_io_hdr header;
	unsigned char recvdata[65535];
	unsigned char cmdbuf[10];
	unsigned int profile, featureslen, ptr, i, feature;
	int result = 0;
	
	memset(&header, 0, sizeof(struct sg_io_hdr));
	header.interface_id = 'S';
	header.flags = SG_FLAG_LUN_INHIBIT | SG_FLAG_DIRECT_IO;
	header.dxfer_direction = SG_DXFER_FROM_DEV;
	header.cmdp = cmdbuf;
	header.cmd_len = sizeof(cmdbuf);
	header.dxferp = recvdata;
	header.dxfer_len = sizeof(recvdata);
	header.sbp = NULL;
	header.mx_sb_len = 0;
	
	memset(&cmdbuf, 0, sizeof(cmdbuf));
	cmdbuf[0] = 0x46;
	cmdbuf[1] = 0;
	cmdbuf[7] = 16;
	cmdbuf[9] = 0;
		
	if ((result = ioctl(fd, SG_IO, &header)) < 0) {
		printf("[-] Failed to send SCSI GET CONFIGURATION command\n");
		return result;
	}
	
	if (header.status == 0) {
		profile = recvdata[6] << 8 | recvdata[7];
		
		device->media_exists = device->media_mo = device->media_cd_rom = device->media_cd_r = device->media_cd_rw = 
		device->media_dvd_rom = device->media_dvd_r = device->media_dvd_ram = device->media_dvd_rw = device->media_dvd_rw_ro = 
		device->media_dvd_rw_seq = device->media_dvd_plus_r = device->media_dvd_plus_rw = device->media_dvd_plus_rw_dl = 
		device->media_dvd_plus_r_dl = device->media_bd = device->media_bd_r = device->media_bd_re = device->media_hddvd = 
		device->media_hddvd_r = device->media_hddvd_rw = device->media_dvd_r_dl_seq = device->media_dvd_r_dl_jmp = 
		device->media_dvd_rw_dl = device->media_hddvd_r_dl = device->media_hddvd_rw_dl =
		device->media_type_cd =	device->media_type_dvd = device->media_type_bd = 0;
		
		if (profile > 0) scsi_profile_set(profile, MMC_PROFILE_MEDIA, device);
		
		result = 1;
	}
								
	if ((featureslen = recvdata[0] << 24 | recvdata[1] << 16 | recvdata[2] << 8 | recvdata[3]) > sizeof(recvdata)) 
		featureslen = sizeof(recvdata);
	
	if (featureslen > 8) {
		for (ptr = 8; ptr + 4 < featureslen; ptr += (4 + recvdata[ptr+3])) {
			if ((feature = recvdata[ptr] << 8 | recvdata[ptr+1]) == 0) {
				device->cap_mo_read = device->cap_cd_rom_read = device->cap_dvd_rw_read = device->cap_dvd_rw_ro_read =
				device->cap_dvd_rw_seq_read = device->cap_dvd_plus_r_read = device->cap_dvd_plus_rw_read = device->cap_dvd_plus_rw_dl_read =
				device->cap_dvd_plus_r_dl_read = device->cap_bd_read = device->cap_bd_r_read = device->cap_bd_re_read =
				device->cap_hddvd_read = device->cap_hddvd_r_read = device->cap_hddvd_rw_read = device->cap_dvd_r_dl_read =
				device->cap_dvd_rw_dl_read = device->cap_hddvd_r_dl_read = device->cap_hddvd_rw_dl_read = 0;
				
				for (i = ptr+4; i+4 <= recvdata[ptr+3]; i += 4) {
					profile = recvdata[i] << 8 | recvdata[i+1];
					scsi_profile_set(profile, MMC_PROFILE_CAP, device);
				}
			}
		}
	}
	
	
	return result;
}

int scsi_mode_sense(int fd, struct optical_device *device)
{
	struct sg_io_hdr header;
	unsigned char recvdata[255];
	unsigned char cmdbuf[10];
	const unsigned int length = 255;
	unsigned short fullsize, pagesize;
	int result = 0;
	
	memset(&header, 0, sizeof(struct sg_io_hdr));
	header.interface_id = 'S';
	header.flags = SG_FLAG_LUN_INHIBIT | SG_FLAG_DIRECT_IO;
	header.dxfer_direction = SG_DXFER_FROM_DEV;
	header.cmdp = cmdbuf;
	header.cmd_len = sizeof(cmdbuf);
	header.dxferp = recvdata;
	header.dxfer_len = sizeof(recvdata);
	header.sbp = NULL;
	header.mx_sb_len = 0;
	
	memset(&cmdbuf, 0, sizeof(cmdbuf));
	cmdbuf[0] = 0x5A;
	cmdbuf[2] = 0x2A;
	cmdbuf[7] = (length >> 8) & 0xff;
	cmdbuf[8] = length & 0xff;
	cmdbuf[9] = 0;
	
	if ((result = ioctl(fd, SG_IO, &header)) < 0) {
		printf("[-] Failed to send SCSI MODE SENSE 10 command\n");
		return result;
	}
	
	if (header.status == 0) {
		fullsize = recvdata[0] * 256 + recvdata[1] + 2;
		
		if ((pagesize = recvdata[9]) >= 20) {
			device->cap_dvd_ram_read = device->cap_dvd_ram_write = device->cap_dvd_r_read = device->cap_dvd_r_write = device->cap_dvd_rom_read = 
			device->cap_simulate = device->cap_cd_rw_read = device->cap_cd_rw_write = device->cap_cd_r_read = device->cap_cd_r_write = 0;
						
			device->cap_dvd_ram_read = (recvdata[10] & 32) ? 1 : 0;
			device->cap_dvd_ram_write = (recvdata[11] & 32) ? 1 : 0;
			device->cap_dvd_r_read = (recvdata[10] & 16) ? 1 : 0;
			device->cap_dvd_r_write = (recvdata[11] & 16) ? 1 : 0;
			device->cap_dvd_rom_read = (recvdata[10] & 8) ? 1 : 0;
			device->cap_simulate = (recvdata[11] & 4) ? 1 : 0;
			device->cap_cd_rw_read = (recvdata[10] & 2) ? 1 : 0;
			device->cap_cd_rw_write = (recvdata[11] & 2) ? 1 : 0;
			device->cap_cd_r_read = (recvdata[10] & 1) ? 1 : 0;
			device->cap_cd_r_write = (recvdata[11] & 1) ? 1 : 0;
			
			device->buffer_size = recvdata[20] * 256 + recvdata[21];
			device->max_read_speed = recvdata[16] * 256 + recvdata[17];
			device->cur_read_speed = recvdata[22] * 256 + recvdata[23];
			device->max_write_speed = recvdata[26] * 256 + recvdata[27];
			device->cur_write_speed = recvdata[28] * 256 + recvdata[29];
		} else {
			device->cap_dvd_rom_read = 1;
			device->cap_simulate = 1;
			device->cap_cd_rw_read = 1;
			device->cap_cd_r_read = 1;
			
			device->buffer_size = 2048;
			device->max_read_speed = 8467;
			device->cur_read_speed = 8467;
			device->max_write_speed = 8467;
			device->cur_write_speed = 8467;
			
			printf("[-] SCSI MODE SENSE 10 2A page not read, using defaults\n");
		}
		
		result = 1;
	}	
	
	return result;
}

int scsi_set_cd_speed(int fd, unsigned short speed, unsigned short limit, struct optical_device *device)
{
	struct sg_io_hdr header;
	unsigned char cmdbuf[12];
	unsigned short rspeed, wspeed;
	int result = 0;
	
	if (!scsi_test_device_writeable(fd)) return 0;
	
	/*if (!scsi_test_unit_ready(fd, 10000)) {
		printf("[-] Device not ready\n");
		return 0;
	}*/

	if (scsi_test_device_media_blank(device)) return 0;
	
	if (device->cur_read_speed == speed) return 1;
	
	memset(&header, 0, sizeof(struct sg_io_hdr));
	header.interface_id = 'S';
	header.flags = SG_FLAG_LUN_INHIBIT | SG_FLAG_DIRECT_IO;
	header.dxfer_direction = SG_DXFER_NONE;
	header.cmdp = cmdbuf;
	header.cmd_len = sizeof(cmdbuf);
	header.dxferp = NULL;
	header.dxfer_len = 0;
	header.sbp = NULL;
	header.mx_sb_len = 0;
	
	if ((speed >= 0) && (speed < MINCDSPEED)) {
		rspeed = MINCDSPEED;
		wspeed = MINCDSPEED;
	} else if (speed >= limit) {
		rspeed = limit;
		wspeed = limit;
	} else {
		rspeed = speed;
		wspeed = speed;
	}
	
	memset(&cmdbuf, 0, sizeof(cmdbuf));
	cmdbuf[0] = 0xbb;
	cmdbuf[2] = rspeed >> 8;
	cmdbuf[3] = rspeed & 0xFF;
	cmdbuf[4] = wspeed >> 8;
	cmdbuf[5] = wspeed & 0xFF;
	
	
	if ((result = ioctl(fd, SG_IO, &header)) < 0) {
		printf("[-] Failed to send SCSI SET CD SPEED command\n");
		return result;
	}
	
	if (header.status == 0) result = 1;
	
	return result;
}

int scsi_set_streaming(int fd, unsigned short speed, unsigned short limit, struct optical_device *device)
{
	struct sg_io_hdr header;
	unsigned char cmdbuf[12];
	unsigned char senddata[28];
	unsigned short rspeed, wspeed;
	int result = 0;
	
	if (!scsi_test_device_writeable(fd)) return 0;
	
	/*if (!scsi_test_unit_ready(fd, 10000)) {
		printf("[-] Device not ready\n");
		return 0;
	}*/

	if (scsi_test_device_media_blank(device)) return 0;
	
	if (device->cur_read_speed == speed) return 1;
	
	memset(&header, 0, sizeof(struct sg_io_hdr));
	header.interface_id = 'S';
	header.flags = SG_FLAG_LUN_INHIBIT | SG_FLAG_DIRECT_IO;
	header.dxfer_direction = SG_DXFER_TO_DEV;
	header.cmdp = cmdbuf;
	header.cmd_len = sizeof(cmdbuf);
	header.dxferp = senddata;
	header.dxfer_len = sizeof(senddata);
	header.sbp = NULL;
	header.mx_sb_len = 0;
	
	if ((speed >= 0) && (speed < MINDVDSPEED)) {
		rspeed = MINDVDSPEED;
		wspeed = MINDVDSPEED;
	} else if (speed >= limit) {
		rspeed = limit;
		wspeed = limit;
	} else {
		rspeed = speed;
		wspeed = speed;
	}
	
	memset(&cmdbuf, 0, sizeof(cmdbuf));
	cmdbuf[0] = 0xb6;
	cmdbuf[10] = 28;
	
	memset(&senddata, 0, sizeof(senddata));
	
	if (rspeed == 0) senddata[0] = 5; else senddata[0] = 1;
	
	if (device->media_end_lba == 0) device->media_end_lba = 2294921;
	
	//Start LBA
	senddata[4] = 0;
	senddata[5] = 0;
	senddata[6] = 0;
	senddata[7] = 0;
	//End LBA
	senddata[8] =  (device->media_end_lba >> 24) & 0xff;
	senddata[9] =  (device->media_end_lba >> 16) & 0xff;
	senddata[10] = (device->media_end_lba >> 8) & 0xff;
	senddata[11] = (device->media_end_lba) & 0xff;
	//Read Size
	senddata[12] = (rspeed >> 24) & 0xff;
	senddata[13] = (rspeed >> 16) & 0xff;
	senddata[14] = (rspeed >> 8) & 0xff;
	senddata[15] =  rspeed & 0xff;
	//Write Size
	senddata[20] = (wspeed >> 24) & 0xff;
	senddata[21] = (wspeed >> 16) & 0xff;
	senddata[22] = (wspeed >> 8) & 0xff;
	senddata[23] =  wspeed & 0xff;
	//Read Time, Write Time
	senddata[18] = senddata[26] = 1000 >> 8;
	senddata[19] = senddata[27] = 1000 & 0xff; 
		
	if ((result = ioctl(fd, SG_IO, &header)) < 0) {
		perror("[-] Failed to send SCSI SET STREAMING command");
		return result;
	}
	
	if (header.status == 0) result = 1;
	
	return result;
}

int scsi_get_capacity(int fd, struct optical_device *device)
{
	struct sg_io_hdr header;
	unsigned char recvdata[8];
	unsigned char cmdbuf[10];
	int result = 0;
	
	memset(&header, 0, sizeof(struct sg_io_hdr));
	header.interface_id = 'S';
	header.flags = SG_FLAG_LUN_INHIBIT | SG_FLAG_DIRECT_IO;
	header.dxfer_direction = SG_DXFER_FROM_DEV;
	header.cmdp = cmdbuf;
	header.cmd_len = sizeof(cmdbuf);
	header.dxferp = recvdata;
	header.dxfer_len = sizeof(recvdata);
	header.sbp = NULL;
	header.mx_sb_len = 0;
	
	memset(&cmdbuf, 0, sizeof(cmdbuf));
	cmdbuf[0] = 0x25;
		
	if ((result = ioctl(fd, SG_IO, &header)) < 0) {
		printf("[-] Failed to send SCSI READ CAPACITY command\n");
		return result;
	}
	
	if (header.status == 0) {
			device->media_end_lba = (recvdata[0] << 24) |
									(recvdata[1] << 16) |
									(recvdata[2] << 8) |
									(recvdata[3]);
			device->media_block_length = (recvdata[4] << 24) |
										(recvdata[5] << 16) |
										(recvdata[6] << 8) |
										(recvdata[7]);
			result = 1;
	}
	
	return result;
}

int scsi_get_disc_information(int fd, struct optical_device *device)
{
	struct sg_io_hdr header;
	unsigned char recvdata[34];
	unsigned char cmdbuf[10];
	const unsigned int length = 34;
	const unsigned int page = 0;
	int result = 0;
	
	memset(&header, 0, sizeof(struct sg_io_hdr));
	header.interface_id = 'S';
	header.flags = SG_FLAG_LUN_INHIBIT | SG_FLAG_DIRECT_IO;
	header.dxfer_direction = SG_DXFER_FROM_DEV;
	header.cmdp = cmdbuf;
	header.cmd_len = sizeof(cmdbuf);
	header.dxferp = recvdata;
	header.dxfer_len = sizeof(recvdata);
	header.sbp = NULL;
	header.mx_sb_len = 0;
	
	memset(&cmdbuf, 0, sizeof(cmdbuf));
	cmdbuf[0] = 0x51;
	cmdbuf[1] = (page & 0x07);
	cmdbuf[7] = (length >> 8) & 0xff;
	cmdbuf[8] = length & 0xff;
	
	
	if ((result = ioctl(fd, SG_IO, &header)) < 0) {
		printf("[-] Failed to send SCSI READ DISC INFORMATION command\n");
		return result;
	}
	
	if (header.status == 0) {
		device->media_sessions_number = (recvdata[9] << 8) | recvdata[4];
		device->media_erasable = (recvdata[2] & 0x10) >> 4;
		device->media_disc_status = recvdata[2] & 0x03;
		device->media_session_state = (recvdata[2] & 0x0C) >> 2;
		device->media_bg_format_status = recvdata[7] & 0x03;
		
		result = 1;
	}
	
	return result;
}

int scsi_read_media_serial_number(int fd, struct optical_device *device)
{
	struct sg_io_hdr header;
	unsigned char recvdata[256];
	unsigned char cmdbuf[12];
	const unsigned int length = 256;
	const unsigned int page = 0x01;
	int result = 0;
	
	memset(&header, 0, sizeof(struct sg_io_hdr));
	header.interface_id = 'S';
	header.flags = SG_FLAG_LUN_INHIBIT | SG_FLAG_DIRECT_IO;
	header.dxfer_direction = SG_DXFER_FROM_DEV;
	header.cmdp = cmdbuf;
	header.cmd_len = sizeof(cmdbuf);
	header.dxferp = recvdata;
	header.dxfer_len = sizeof(recvdata);
	header.sbp = NULL;
	header.mx_sb_len = 0;
	
	memset(&cmdbuf, 0, sizeof(cmdbuf));
	cmdbuf[0] = 0xab;
	cmdbuf[1] = 0x01;
	cmdbuf[6] = (length >> 24) & 0xff;
    cmdbuf[7] = (length >> 16) & 0xff;
    cmdbuf[8] = (length >> 8) & 0xff;
    cmdbuf[9] = length & 0xff;
	
	if ((result = ioctl(fd, SG_IO, &header)) < 0) {
		printf("[-] Failed to send SCSI READ MEDIA SERIAL NUMBER command\n");
		return result;
	}
	
	if (header.status == 0) {
		int i;
		for (i=0; i<256; i++) {
			printf("0x%2x", recvdata[i]);
		}
		printf("\n");
		
		result = 1;
	}
	
	return result;
}

int scsi_get_event_status_notification(int fd, int pollflag, struct optical_device *device)
{
	struct sg_io_hdr header;
	unsigned char recvdata[8];
	unsigned char cmdbuf[10];
	int result = 0;
	
	memset(&header, 0, sizeof(struct sg_io_hdr));
	header.interface_id = 'S';
	header.flags = SG_FLAG_LUN_INHIBIT | SG_FLAG_DIRECT_IO;
	header.dxfer_direction = SG_DXFER_FROM_DEV;
	header.cmdp = cmdbuf;
	header.cmd_len = sizeof(cmdbuf);
	header.dxferp = recvdata;
	header.dxfer_len = sizeof(recvdata);
	header.sbp = NULL;
	header.mx_sb_len = 0;
	
	memset(&cmdbuf, 0, sizeof(cmdbuf));
	cmdbuf[0] = 0x4A;
	cmdbuf[1] = pollflag & 1;
	cmdbuf[2] = 0;
	cmdbuf[3] = 0;
	cmdbuf[4] = 0x10;
	cmdbuf[5] = 0;
	cmdbuf[6] = 0;
	cmdbuf[7] = 0;
	cmdbuf[8] = 8;
	cmdbuf[9] = 0;
	
	if ((result = ioctl(fd, SG_IO, &header)) < 0) {
		printf("[-] Failed to send SCSI GET EVENT STATUS NOTIFICATION command\n");
		return result;
	}
	
	if (header.status == 0) {
		device->event_eject_request = device->event_new_media = device->event_media_removal = device->event_media_changed =
		device->event_bgformat_completed = device->event_bgformat_restarted = device->event_tray_open = device->event_media_present = 0;
		if ((recvdata[1] > 4) && ((((recvdata[2] >> 7) & 0x01) == 0x00) && ((recvdata[2] & 0x07) == 0x04))) {
			switch (recvdata[4] & 0x0f) {
				case 0x00: 
					break;
				case 0x01:
					device->event_eject_request = 1;
					break;
				case 0x02:
					device->event_new_media = 1;
					break;
				case 0x03:
					device->event_media_removal = 1;
					break;
				case 0x04:
					device->event_media_changed = 1;
					break;
				case 0x05:
					device->event_bgformat_completed = 1;
					break;
				case 0x06:
					device->event_bgformat_restarted = 1;
					break;
				default:
					break;
			}
			device->event_tray_open = (recvdata[5] & 0x01);
			device->event_media_present = ((recvdata[5] >> 1) & 0x0f);
			//printf("--> 0x%2x 0x%2x 0x%2x 0x%2x 0x%2x 0x%2x 0x%2x 0x%2x\n", recvdata[0], recvdata[1], recvdata[2], recvdata[3], recvdata[4], recvdata[5], recvdata[6], recvdata[7]);
			result = 1;
		} else {
			result = -1;
		}
		
	}
	
	return result;
}
