/*
 *      scsi.h
 *      
 *      Copyright 2011 Alex <alex@linuxonly.ru>
 *      
 *      This program is free software: you can redistribute it and/or modify
 *      it under the terms of the GNU General Public License as published by
 *      the Free Software Foundation; either version 3 of the License, or
 *      (at your option) any later version.
 *      
 *      This program is distributed in the hope that it will be useful,
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of
 *      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *      GNU General Public License for more details.
 *      
 *      You should have received a copy of the GNU General Public License
 *      along with this program. If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef __SCSI_H__
#define __SCSI_H__

#include <math.h>

#define CDSPEED(x)  ( (unsigned int)(round((x)/177.0)) )
#define DVDSPEED(x) ( (unsigned int)(round((x)/1385.0)) )
#define BDSPEED(x)  ( (unsigned int)(round((x)/5540.0)) )

#define CDBYTESTREAM(x)  ( (unsigned int)((x)*177) )
#define DVDBYTESTREAM(x) ( (unsigned int)((x)*1385) )
#define BDBYTESTREAM(x)  ( (unsigned int)((x)*5540) )

#define MINCDSPEED   177
#define MINDVDSPEED  1386
#define MINBDSPEED   5540

#define MAXCDSPEED   177*56
#define MAXDVDSPEED  1386*24
#define MAXBDSPEED   5540*12

enum {
	MMC_PROFILE_MEDIA = 0,
	MMC_PROFILE_CAP
};

int scsi_profile_set(int profile, int action, struct optical_device *device);
int scsi_test_unit_ready(int fd, unsigned int timeout);
int scsi_test_device_writeable(int fd);
int scsi_test_device_media_blank(struct optical_device *device);
int scsi_inquiry(int fd, int page_code, int evpd, struct optical_device *device);
int scsi_get_configuration(int fd, struct optical_device *device);
int scsi_mode_sense(int fd, struct optical_device *device);
int scsi_set_cd_speed(int fd, unsigned short speed, unsigned short limit, struct optical_device *device);
int scsi_set_streaming(int fd, unsigned short speed, unsigned short limit, struct optical_device *device);
int scsi_get_capacity(int fd, struct optical_device *device);
int scsi_get_disc_information(int fd, struct optical_device *device);

int scsi_read_media_serial_number(int fd, struct optical_device *device);
int scsi_get_event_status_notification(int fd, int pollflag, struct optical_device *device);

#endif /* __SCSI_H__ */
