/*
 *      settings.h
 *      
 *      Copyright 2011 Alex <alex@linuxonly.ru>
 *      
 *      This program is free software: you can redistribute it and/or modify
 *      it under the terms of the GNU General Public License as published by
 *      the Free Software Foundation; either version 3 of the License, or
 *      (at your option) any later version.
 *      
 *      This program is distributed in the hope that it will be useful,
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of
 *      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *      GNU General Public License for more details.
 *      
 *      You should have received a copy of the GNU General Public License
 *      along with this program. If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef __SETTINGS_H__
#define __SETTINGS_H__

#include <grp.h>
#include <pwd.h>
#include <gdbm.h>

#include "cdsd.h"

#define SETTINGS_SOCKET_FILE         "/var/run/cdsd.socket"
#define SETTINGS_SOCKET_LISTENQ      10

#define SETTINGS_DB_ENTRY            "<dev cd=\"%u\" dvd=\"%u\" bd=\"%u\" poll=\"%u\">%s</dev>\n"

#define SETTINGS_FILE_DIR            "/var/lib/cdsd"
#define SETTINGS_DB_FILENAME         "/var/lib/cdsd/devices.gdbm"
#define SETTINGS_CONFIG_FILE_NAME    "/etc/cdsd.conf"
#define SETTINGS_LOCK_FILE_NAME      "/var/run/cdsd.pid"


#define CDSD_LIST_FORMAT       "%s8%s16%s16%i%i%i%i%i%i%i"
/*< s - vendor, s - product, s - filename, i - id, i - media, i - capabilities, i - cdspeed, i - dvdspeed, i - bdspeed, i - polling >*/
#define CDSD_DEVICE_FORMAT     "%i%i%i%i%i"
/*i - id, i - cdspeed, i - dvdspeed, i - bdspeed, i - polling*/
#define CDSD_REQUEST_FORMAT    "%i"
/*i - id*/
#define CDSD_INFO_FORMAT       "%s8%s16%s4%s20%h%h%h%h%h%b%b%b%b%b"
/*s - vendor, s - product, s - version, s - serial, h - buffer size, h - max read speed, h - cur read speed, h - max write speed, h - cur write speed,
 *b - media exists, b - media erasable, b - media is cd, b - media is dvd, b - media is bd*/
#define CDSD_CAPABILITIES_FORMAT       "%b%b%b%b%b%b%b%b%b%b%b%b%b%b%b%b%b%b%b%b%b"
/*b - cdrom, b - cd_r, b - cd_rw, b - dvd_rom, b - dvd_ram, b - dvd_r, b - dvd_r_dl, b - dvd_rw, b - dvd_rw_dl, b - dvd_plus_r, b - dvd_plus_r_dl,
 *b - dvd_plus_rw, b - dvd_plus_rw_dl, b - bd_rom, b - bd_r, b - bd_re, b - hddvd_rom, b - hddvd_r, b - hddvd_r_dl, b - hddvd_rw, b - hddvd_rw_dl;*/

#define CDSD_COMMAND_UNKNOWN        0x00
#define CDSD_COMMAND_LIST           0x01
#define CDSD_COMMAND_SET            0x02
#define CDSD_COMMAND_INFO           0x03
#define CDSD_COMMAND_CAPABILITIES   0x04

#define CDSD_RESULT_ERROR           0x00
#define CDSD_RESULT_SUCCESS         0x01

struct _cmd {
	unsigned short command;
	unsigned short result;
	unsigned short error;
};

int settings_parse_config_file(const char *filename, const char *section, void *userdata, int(*callback(char *parameter, char *argument, void *userdata)));
int settings_server_connection_create(void);
void settings_server_connection_close(int sockfd);
int settings_read_credentials(int sockfd, struct _cmd *cmd, char *data, size_t datasize, struct ucred *credentials);
int settings_client_connection_create(void);
int settings_send_credentials(int sockfd, struct _cmd *cmd, char *data, size_t datasize);
gid_t settings_get_group_id(char *groupname);
int settings_user_in_group(int uid, int gid);
char *settings_string_get_fragment(char *str, char *param, char *term, char *buffer, ssize_t bufsize);
unsigned int settings_string_get_uint(char *str, char *param, char *term);
long int settings_string_get_int(char *str, char *param, char *term);

GDBM_FILE settings_db_open(char *filename);
void settings_db_close(GDBM_FILE dbf);
int settings_db_fetch(GDBM_FILE dbf, struct optical_device *device);
int settings_db_store(GDBM_FILE dbf, struct optical_device *device);

#endif /* __SETTINGS_H__ */
